# -*- coding: utf-8 -*-
"""
MIT License

Copyright (c) 2021 Leonardo Eiji Tamayose, Guilherme Ferrari Fortino

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import numpy as np
import pandas as pd
from PyQt5.QtGui import QGuiApplication
from PyQt5.QtCore import QObject, QJsonValue, QUrl, QVariant, pyqtSignal, pyqtSlot
from scipy.odr import ODR, Model as SciPyModel, RealData
from lmfit.models import ExpressionModel
from lmfit import Parameters
from copy import deepcopy
from io import StringIO
import re


class Model(QObject):
    """
    Class used for fit.
    """
    # Signals
    fillDataTable = pyqtSignal(str,
                               str,
                               str,
                               str,
                               str,
                               str,
                               arguments=["x", "y", "sy", "sx", "filename"])
    fillParamsTable = pyqtSignal(str,
                                 float,
                                 float,
                                 arguments=["param", "value", "uncertainty"])
    writeInfos = pyqtSignal(str, arguments="expr")
    uploadData = pyqtSignal(QVariant, str, arguments=["data", "fileName"])

    def __init__(self, messageHandler):
        super().__init__()
        pd.set_option("display.expand_frame_repr", False)
        self._msg_handler = messageHandler
        self._data = None
        self._data_json = None
        self._exp_model = ""
        self._indVar = "x"
        self._model = None
        self._report_fit = ""
        self._mat_corr = ""
        self._mat_cov = ""
        self._dict_param = {}
        self._result = None
        self._coef = []
        self._par_var = []
        self._params = Parameters()
        self._dict = {}
        self._dict2 = {}
        self._p0 = None
        self.xmin_adj = 0.
        self.xmax_adj = 0.
        self._mode = 0
        self._has_data = False
        self._isvalid = False
        self._has_sx = True
        self._has_sy = True
        self._indices = []

    def __str__(self):
        return self._report_fit

    @pyqtSlot(QJsonValue)
    def loadDataTable(self, data=None):
        """ Getting data from table. """
        df = pd.DataFrame.from_records(data,
                                       columns=["x", "y", "sy", "sx", "bool"])

        # Removing not chosen rows
        df = df[df["bool"] == 1]
        del df["bool"]
        uniqueSi = df["sy"].unique().astype(float)
        if 0. in uniqueSi:
            if len(uniqueSi) > 1:
                self._msg_handler.raise_warn(
                    "Um valor nulo foi encontrado nas incertezas em y, removendo coluna de sy."
                )
            self._has_sy = False
        uniqueSi = df["sx"].unique().astype(float)
        if 0. in uniqueSi:
            if len(uniqueSi) > 1:
                self._msg_handler.raise_warn(
                    "Um valor nulo foi encontrado nas incertezas em x, removendo coluna de sx."
                )
            self._has_sx = False

        self._data_json = deepcopy(df)

        # Turn everything into number (str -> number)
        df = df.astype(float)
        self._data = deepcopy(df)
        self._has_data = True

    @pyqtSlot()
    def loadDataClipboard(self):
        """Pega a tabela de dados do Clipboard."""
        # Instantiating clipboard
        clipboard = QGuiApplication.clipboard()
        clipboardText = clipboard.mimeData().text()
        # try:
        # Creating a dataframe from the string
        df = pd.read_csv(StringIO(clipboardText),
                         sep="\t",
                         header=None,
                         dtype=str).dropna(how="all").replace(np.nan, "0")
        # Replacing all commas for dots
        for i in df.columns:
            df[i] = [x.replace(",", ".") for x in df[i]]
            df[i] = df[i].astype(str)
        self.load_data(df=df)

    @pyqtSlot(str)
    def load_data(self, data_path="", df=None, df_array=None):
        """Loads the data from a given path or from a given dataframe."""

        # Name of the loaded file
        fileName = "Dados Carregados do Projeto"

        # If no dataframe passed, loading data from the given path
        if len(data_path) > 0:
            # Loading from .csv or (.txt and .tsv)
            data_path = QUrl(data_path).toLocalFile()
            if data_path[-3:] == "csv":
                try:
                    df = pd.read_csv(data_path,
                                     sep=",",
                                     header=None,
                                     dtype=str).dropna(how="all").replace(
                                         np.nan, "0")
                except pd.errors.ParserError:
                    self._msg_handler.raise_error(
                        "Separação de colunas de arquivos csv são com vírgula (','). Rever dados de entrada."
                    )
                    return None
            else:
                try:
                    df = pd.read_csv(data_path,
                                     sep="\t",
                                     header=None,
                                     dtype=str).dropna(how="all").replace(
                                         np.nan, "0")
                except pd.errors.ParserError:
                    self._msg_handler.raise_error(
                        "Separação de colunas de arquivos txt e tsv são com tab. Rever dados de entrada."
                    )
                    return None
            # Getting file name
            fileName = data_path.split("/")[-1]
        elif df is None:
            df = pd.DataFrame.from_records(
                df_array, columns=["x", "y", "sy", "sx", "bool"])
            del df["bool"]
            uniqueSi = df["sy"].unique().astype(float)
            if 0. in uniqueSi:
                if len(uniqueSi) > 1:
                    self._msg_handler.raise_warn(
                        "Um valor nulo foi encontrado nas incertezas em y, removendo coluna de sy."
                    )
                self._has_sy = False
            uniqueSi = df["sx"].unique().astype(float)
            if 0. in uniqueSi:
                if len(uniqueSi) > 1:
                    self._msg_handler.raise_warn(
                        "Um valor nulo foi encontrado nas incertezas em x, removendo coluna de sx."
                    )
                self._has_sx = False

        # Saving the dataframe in the class
        self._data_json = deepcopy(df)

        # Applying some filters over the df
        for i in df.columns:
            # Replacing comma for dots
            df[i] = [x.replace(",", ".") for x in df[i]]
            self._data_json[i] = [
                x.replace(",", ".") for x in self._data_json[i]
            ]
            try:
                df[i] = df[i].astype(float)
            except ValueError:
                self._msg_handler.raise_error(
                    "A entrada de dados só permite entrada de números. Rever arquivo de entrada."
                )
                return None

        self._has_sx = True
        self._has_sy = True
        self._mode = len(df.columns) - 2

        # Naming columns
        if self._mode == -1:
            self._has_sy = not self._has_sy
            self._has_sx = not self._has_sx
            df["x"] = np.arange(len(df), dtype=float)
            self._data_json = deepcopy(df.astype(str))
            self._data_json.columns = ["y", "x"]
            df["sy"] = 0.
            df["sx"] = 0.
        elif self._mode == 0:
            self._has_sy = not self._has_sy
            self._has_sx = not self._has_sx
            self._data_json.columns = ["x", "y"]
            df["sy"] = 0.
            df["sx"] = 0.
        elif self._mode == 1:
            self._has_sx = not self._has_sx
            self._data_json.columns = ["x", "y", "sy"]
            df["sx"] = 0.
        else:
            try:
                self._data_json.columns = ["x", "y", "sy", "sx"]
                uniqueSi = self._data_json["sy"].unique().astype(float)
                if 0. in uniqueSi:
                    if len(uniqueSi) > 1:
                        self._msg_handler.raise_warn(
                            "Um valor nulo foi encontrado nas incertezas em y, removendo coluna de sy."
                        )
                    self._has_sy = False
                uniqueSi = self._data_json["sx"].unique().astype(float)
                if 0. in uniqueSi:
                    if len(uniqueSi) > 1:
                        self._msg_handler.raise_warn(
                            "Um valor nulo foi encontrado nas incertezas em x, removendo coluna de sx."
                        )
                    self._has_sx = False
            except ValueError:
                self._msg_handler.raise_error(
                    "Há mais do que 4 colunas. Rever entrada de dados.")
                return None

        self._data = deepcopy(df)
        self._has_data = True
        self.uploadData.emit(self._data_json.to_dict(orient="list"), fileName)

    def set_p0(self, p0):
        ''' Coloca os chutes iniciais. '''
        self._p0 = p0.replace(" ", "").split(",")

    def set_expression(self, exp="", varInd="x"):
        """ Set new expression to model. """
        self._exp_model = exp
        self._indVar = varInd

    def fit(self, **kargs):
        ''' Interpretador de qual ajuste deve ser feito. '''
        wsx = kargs.pop("wsx", True)
        wsy = kargs.pop("wsy", True)

        # Getting Model
        try:
            self._model = ExpressionModel(self._exp_model +
                                          " + 0*%s" % self._indVar,
                                          independent_vars=[self._indVar])
        except ValueError:
            self._msg_handler.raise_error(
                "Expressão de ajuste escrita de forma errada. Rever função de ajuste."
            )
            return None
        except SyntaxError:
            self._msg_handler.raise_error(
                "Expressão de ajuste escrita de forma errada. Rever função de ajuste."
            )
            return None
        # Getting coefficients
        self._coef = [i for i in self._model.param_names]
        # Data
        x, y, sy, sx = self.data
        self._indices = np.arange(len(self._data.index))
        if self.xmin != self.xmax:
            self._indices = np.where((self.xmin <= self._data["x"])
                               & (self.xmax >= self._data["x"]))[0]
        x, y, sy, sx = x.iloc[self._indices].to_numpy(), y.iloc[self._indices].to_numpy(
        ), sy.iloc[self._indices].to_numpy(), sx.iloc[self._indices].to_numpy()
        data = None
        if self._has_sy and self._has_sx:  # Caso com as duas incs
            if wsx == True and wsy == True:
                self.__fit_lm_wy(x, y)
                if (self._result is None) == False:
                    self.__set_param_values_lm_special(x)
                    self.__set_report_lm_special(x)
                else:
                    return None
            elif wsx:
                self.__fit_lm(x, y, sy)
                if (self._result is None) == False:
                    self.__set_param_values_lm(x)
                    self.__set_report_lm(x)
                else:
                    return None
            elif wsy:
                self.__fit_ODR_special(x, y, sx)
                if (self._result is None) == False:
                    self.__set_param_values_ODR(x)
                    self.__set_report_ODR(x)
                else:
                    return None
            else:
                data = RealData(x, y, sx=sx, sy=sy)
                self.__fit_ODR(data)
                if (self._result is None) == False:
                    self.__set_param_values_ODR(x)
                    self.__set_report_ODR(x)
                else:
                    return None
        elif self._has_sy:  # Caso com a incerteza só em y
            if wsy:
                self.__fit_lm_wy(x, y)
                if (self._result is None) == False:
                    self.__set_param_values_lm_special(x)
                    self.__set_report_lm_special(x)
                else:
                    return None
            else:
                self.__fit_lm(x, y, sy)
                if (self._result is None) == False:
                    self.__set_param_values_lm(x)
                    self.__set_report_lm(x)
                else:
                    return None
        elif self._has_sx:  # Caso com a incerteza só em x
            if wsx:
                self.__fit_lm_wy(x, y)
                if (self._result is None) == False:
                    self.__set_param_values_lm_special(x)
                    self.__set_report_lm_special(x)
                else:
                    return None
            else:
                self.__fit_ODR_special(x, y, sx)
                if (self._result is None) == False:
                    self.__set_param_values_ODR(x)
                    self.__set_report_ODR(x)
                else:
                    return None
        else:  # Caso sem incertezas
            self.__fit_lm_wy(x, y)
            if (self._result is None) == False:
                self.__set_param_values_lm_special(x)
                self.__set_report_lm_special(x)
            else:
                return None
        params = self.get_params()
        keys = list(params.keys())
        for i in range(len(keys)):
            self.fillParamsTable.emit(keys[i], params[keys[i]][0],
                                      params[keys[i]][1])
        self.writeInfos.emit(self._report_fit)

    def __make_parameters_lm(self):
        '''Constrói os parâmetros para ajuste com o lmfit.'''
        self._params = Parameters()
        if self._p0 is None:
            for i in range(len(self._coef)):
                self._params.add(self._coef[i], 1.)
        else:
            coefs = {c: [1, True, -np.inf, np.inf] for c in self._coef}
            coefs_2 = {c: False
                       for c in self._coef
                       }  # Para evitar de substituir atribuição de parâmetros
            for i in range(len(self._coef)):
                try:
                    res = re.match(
                        r"((?P<parameter>.+?)=)?(?P<value>[\d\-\.\@]+)(\[((?P<lim_inf>.+?)?;(?P<lim_sup>.+?))\])?",
                        self._p0[i]).groupdict()
                    if res["parameter"] is not None:
                        valor = res["value"].replace("@", "")
                        var = res["parameter"]
                        if var in coefs:
                            lim_inf = -np.inf if res[
                                "lim_inf"] is None else float(res["lim_inf"])
                            lim_sup = np.inf if res[
                                "lim_sup"] is None else float(res["lim_sup"])
                            coefs[var] = [
                                float(valor), not "@" in res["value"], lim_inf,
                                lim_sup
                            ]
                            coefs_2[var] = True
                    else:
                        if coefs_2[self._coef[i]] == False and self._coef[
                                i] in coefs:
                            lim_inf = -np.inf if res[
                                "lim_inf"] is None else float(res["lim_inf"])
                            lim_sup = np.inf if res[
                                "lim_sup"] is None else float(res["lim_sup"])
                            coefs[self._coef[i]] = [
                                float(res["value"].replace("@", "")),
                                not "@" in res["value"], lim_inf, lim_sup
                            ]
                            coefs_2[self._coef[i]] = True
                except:
                    if coefs_2[self._coef[i]] == False:
                        coefs[self._coef[i]] = [1, True, -np.inf, np.inf]
                        coefs_2[self._coef[i]] = True
            for nome in coefs.keys():
                self._params.add(nome,
                                 coefs[nome][0],
                                 vary=coefs[nome][1],
                                 min=coefs[nome][2],
                                 max=coefs[nome][3])

    def __make_parameters_odr(self):
        '''Constrói os parâmetros para ajuste com o ODR.'''
        pi = [1.] * len(self._coef)
        fixed = [1] * len(self._coef)
        arr_lim_inf = [-np.inf] * len(self._coef)
        arr_lim_sup = [np.inf] * len(self._coef)
        aux = {c: i for i, c in enumerate(self._coef)}
        if self._p0 is None:
            pass
        else:
            coefs = {c: [1, True, -np.inf, np.inf] for c in self._coef}
            coefs_2 = {c: False
                       for c in self._coef
                       }  # Para evitar de substituir atribuição de parâmetros
            for i in range(len(self._coef)):
                try:
                    res = re.match(
                        r"((?P<parameter>.+?)=)?(?P<value>[\d\-\.\@]+)(\[((?P<lim_inf>.+?)?;(?P<lim_sup>.+?))\])?",
                        self._p0[i]).groupdict()
                    if res["parameter"] is not None:
                        valor = res["value"].replace("@", "")
                        var = res["parameter"]
                        lim_inf = -np.inf if res["lim_inf"] is None else float(
                            res["lim_inf"])
                        lim_sup = np.inf if res["lim_sup"] is None else float(
                            res["lim_sup"])
                        if var in coefs:
                            coefs[var] = [
                                float(valor), not "@" in res["value"], lim_inf,
                                lim_sup
                            ]
                            coefs_2[var] = True
                    else:
                        if coefs_2[self._coef[i]] == False and self._coef[
                                i] in coefs:
                            lim_inf = -np.inf if res[
                                "lim_inf"] is None else float(res["lim_inf"])
                            lim_sup = np.inf if res[
                                "lim_sup"] is None else float(res["lim_sup"])
                            coefs[self._coef[i]] = [
                                float(res["value"].replace("@", "")),
                                not "@" in res["value"], lim_inf, lim_sup
                            ]
                            coefs_2[self._coef[i]] = True
                except:
                    if coefs_2[self._coef[i]] == False:
                        coefs[self._coef[i]] = [1, True, -np.inf, np.inf]
                        coefs_2[self._coef[i]] = True
            for i, nome in enumerate(self._coef):
                pi[aux[nome]] = coefs[nome][0]
                fixed[aux[nome]] = coefs[nome][1]
                arr_lim_inf[aux[nome]] = coefs[nome][2]
                arr_lim_sup[aux[nome]] = coefs[nome][3]
        self._par_var = []
        for i, parametro in enumerate(self._coef):
            if fixed[i] > 0:
                self._par_var.append(parametro)
        return pi, fixed, arr_lim_inf, arr_lim_sup

    def __fit_ODR(self, data):
        '''Fit com ODR.'''
        pi, fixed, lim_inf, lim_sup = self.__make_parameters_odr()

        def f(a, x):
            param = Parameters()
            for i in range(len(a)):
                param.add(self._model.param_names[i],
                          value=a[i],
                          vary=fixed[i],
                          min=lim_inf[i],
                          max=lim_sup[i])
            return eval("self._model.eval(%s=x, params=param)" % self._indVar,
                        {
                            "x": x,
                            "param": param,
                            "self": self
                        })

        model = SciPyModel(f)
        try:
            myodr = ODR(data, model, beta0=pi, maxit=200, ifixb=fixed)
            self._result = myodr.run()
        except TypeError:
            self._msg_handler.raise_error(
                "Expressão de ajuste escrita de forma errada. Rever função de ajuste."
            )
            return None

    def __fit_ODR_special(self, x_orig, y, sx):
        '''Fit com ODR quando só há incertezas em x.'''
        pi, fixed, lim_inf, lim_sup = self.__make_parameters_odr()

        def f(a, x):
            param = Parameters()
            for i in range(len(a)):
                param.add(self._model.param_names[i],
                          value=a[i],
                          vary=fixed[i],
                          min=lim_inf[i],
                          max=lim_sup[i])
            return eval("self._model.eval(%s=x, params=param)" % self._indVar,
                        {
                            "x": x,
                            "param": param,
                            "self": self
                        })

        # data  = RealData(x, y, sx = sx)
        # model = SciPyModel(f)
        # try:
        #     myodr = ODR(data, model, beta0 = pi, maxit = 40)
        #     self._result = myodr.run()
        # except TypeError:
        #     self._msg_handler.raise_error("Expressão de ajuste escrita de forma errada. Rever função de ajuste.")
        #     self._result = None
        #     return None
        # self._params = Parameters()
        # for i in range(len(self._coef)):
        #     self._params.add(self._coef[i], self._result.beta[i])
        # sy = np.zeros(len(self._data["x"]), dtype = float)
        # for i, x in enumerate(self._data["x"]):
        #     x_var = np.array([x + self._data["sx"].iloc[i], x - self._data["sx"].iloc[i]])
        #     y_prd = eval("self._model.eval(%s = x, params = self._params)"%self._indVar, None,
        # {"x": x, "self": self})
        #     y_var = eval("self._model.eval(%s = x_var, params = self._params)"%self._indVar, None,
        # {"x_var": x_var, "self": self})
        #     sy[i] = np.abs(y_var - y_prd).mean()
        # sy = sy.astype(float)/1000
        x = np.copy(x_orig)
        sy = np.array([1e-50] * len(x), dtype=float)
        data = RealData(x, y, sx=sx, sy=sy)
        model = SciPyModel(f)
        try:
            myodr = ODR(data, model, beta0=pi, maxit=100, ifixb=fixed)
            self._result = myodr.run()
        except TypeError as e:
            print(e)
            self._msg_handler.raise_error(
                "Expressão de ajuste escrita de forma errada. Rever função de ajuste."
            )
            self._result = None
            return None
        sy = np.zeros(len(self._data["x"]), dtype=float)
        for i, x in enumerate(self._data["x"]):
            x_var = np.array(
                [x + self._data["sx"].iloc[i], x - self._data["sx"].iloc[i]])
            y_prd = eval(
                "self._model.eval(%s = x, params = self._params)" %
                self._indVar, None, {
                    "x": x,
                    "self": self
                })
            y_var = eval(
                "self._model.eval(%s = x_var, params = self._params)" %
                self._indVar, None, {
                    "x_var": x_var,
                    "self": self
                })
            sy[i] = np.abs(y_var - y_prd).mean()
        x_var = self._data["x"]
        self._result.sum_square = np.sum(((eval(
            "self._model.eval(%s = x_var, params = self._params)" %
            self._indVar, None, {
                "x_var": x_var.to_numpy(),
                "self": self
            }) - self._data["y"].to_numpy()) / sy)**2)

    def __fit_lm(self, x, y, sy):
        '''Fit com MMQ.'''
        self.__make_parameters_lm()
        try:
            self._result = eval(
                "self._model.fit(data = y, %s = x, weights = 1/sy, params = params, scale_covar=False, max_nfev = 250)"
                % self._indVar, None, {
                    "y": y,
                    "x": x,
                    "params": self._params,
                    "self": self,
                    "sy": sy
                })
        except ValueError:
            self._msg_handler.raise_error(
                "A função ajustada gera valores não numéricos, rever ajuste e/ou parâmetros inciais."
            )
            return None
        except TypeError:
            self._msg_handler.raise_error(
                "A função ajustada possui algum termo inválido, rever ajuste e/ou parâmetros inciais."
            )
            return None
        if self._result.covar is None:
            self._msg_handler.raise_error(
                "A função ajustada não convergiu, rever ajuste e/ou parâmetros inciais."
            )
            self._result = None
            return None

    def __fit_lm_wy(self, x, y):
        '''Fit com MMQ quando não há incertezas.'''
        self.__make_parameters_lm()
        try:
            self._result = eval(
                "self._model.fit(data = y, %s = x, params = self._params, scale_covar=False, max_nfev = 250)"
                % self._indVar, None, {
                    "y": y,
                    "x": x,
                    "self": self
                })
        except ValueError:
            self._msg_handler.raise_error(
                "A função ajustada gera valores não numéricos, rever ajuste e/ou parâmetros inciais."
            )
            return None
        except TypeError:
            self._msg_handler.raise_error(
                "A função ajustada possui algum termo inválido, rever ajuste e/ou parâmetros inciais."
            )
            return None
        if self._result.covar is None:
            self._msg_handler.raise_error(
                "A função ajustada não convergiu, rever ajuste e/ou parâmetros inciais."
            )
            self._result = None
            return None

    def get_params(self):
        '''Retorna um dicionário onde as keys são os parâmetros e que retornam uma lista com [valor, incerteza].'''
        return self._dict

    def __set_param_values_lm(self, x):
        '''Constrói o dicionário e o Parameters dos valores do ajuste.'''
        self._dict.clear()
        self._dict_param.clear()
        self._params = Parameters()
        self._par_var = []
        for i in list(self._result.params.keys()):
            if self._result.params[i].vary:
                self._par_var.append(i)
        for i, param_name in enumerate(self._par_var):
            self._params.add(param_name, self._result.values[param_name])
            # self._dict.update({self._coef[i]: [self._result.values[self._coef[i]], np.sqrt(self._result.covar[i, i])*inc_cons]})
            self._dict.update({
                param_name: [
                    self._result.values[param_name],
                    np.sqrt(self._result.covar[i, i])
                ]
            })
            self._dict_param.update({
                param_name: [
                    self._result.values[param_name],
                    np.sqrt(self._result.covar[i, i])
                ]
            })

    def __set_param_values_lm_special(self, x):
        '''Constrói o dicionário e o Parameters dos valores do ajuste, quando não há incertezas.'''
        self._dict.clear()
        self._dict2.clear()
        self._dict_param.clear()
        self._params = Parameters()
        ngl = len(x) - self._result.nvarys
        inc_cons = np.sqrt(self._result.chisqr / ngl) if ngl > 0 else 1
        self._par_var = []
        for i in list(self._result.params.keys()):
            if self._result.params[i].vary:
                self._par_var.append(i)
        for i, param_name in enumerate(self._par_var):
            self._params.add(param_name, self._result.values[param_name])
            self._dict.update({
                param_name: [
                    self._result.values[param_name],
                    np.sqrt(self._result.covar[i, i]) * inc_cons
                ]
            })
            self._dict2.update({
                param_name: [
                    self._result.values[param_name],
                    np.sqrt(self._result.covar[i, i])
                ]
            })
            self._dict_param.update({
                param_name: [
                    self._result.values[param_name],
                    np.sqrt(self._result.covar[i, i]) * inc_cons
                ]
            })

    def __set_param_values_ODR(self, x):
        '''Constrói o dicionário e o Parameters dos valores do ajuste.'''
        self._dict.clear()
        self._dict_param.clear()
        self._params = Parameters()
        for i in range(len(self._coef)):
            self._params.add(self._coef[i], self._result.beta[i])
            self._dict.update({
                self._coef[i]:
                [self._result.beta[i],
                 np.sqrt(self._result.cov_beta[i, i])]
            })
            self._dict_param.update({self._coef[i]: self._dict[self._coef[i]]})

    def __set_report_lm(self, x):
        '''Constrói a string com os resultados.'''
        self._report_fit = ""
        self._report_fit += "\nAjuste: y = %s\n" % self._exp_model
        self._report_fit += "\nNGL  = %d" % (len(x) - self._result.nvarys)
        self._report_fit += "\nChi² = %f\n\n" % self._result.chisqr
        self._report_fit += self.params_print()
        self._report_fit += "\n"
        self._mat_cov = self._result.covar
        lista = list(self._params.keys())
        matriz_corr = np.zeros(
            (len(self._result.covar), len(self._result.covar)))
        z = range(len(matriz_corr))
        for i in z:
            for j in z:
                matriz_corr[i, j] = self._result.covar[i, j] / (
                    self._dict[lista[i]][1] * self._dict[lista[j]][1])
        self._mat_corr = matriz_corr.round(3)
        self._report_fit += "\nMatriz de correlação:\n\n" + self.matprint(
            self._mat_corr, ".3f") + "\n"
        self._report_fit += "Matriz de covariância:\n\n" + self.matprint(
            self._result.covar, fmt = ".3e") + "\n\n"
        self._isvalid = True

    def __set_report_lm_special(self, x):
        '''Constrói a string com os resultados, neste caso quando não há incertezas.'''
        ngl = len(x) - self._result.nvarys
        inc_considerada = np.sqrt(self._result.chisqr / ngl) if ngl > 0 else 0
        inc_considerada_q = inc_considerada**2
        self._report_fit = ""
        self._report_fit += "\nAjuste: y = %s\n" % self._exp_model
        self._report_fit += "\nNGL  = %d" % (ngl)
        self._report_fit += "\nSomatória dos resíduos absolutos ao quadrado = %f\n" % self._result.chisqr
        self._report_fit += "Incerteza considerada = %f\n\n" % inc_considerada
        try:
            self._mat_cov = self._result.covar * inc_considerada_q
            self._report_fit += self.params_print2(inc_considerada)
            self._report_fit += "\n"
            lista = list(self._params.keys())
            matriz_corr = np.zeros(
                (len(self._result.covar), len(self._result.covar)))
            z = range(len(matriz_corr))
            for i in z:
                for j in z:
                    matriz_corr[i, j] = self._result.covar[i, j] / (
                        self._dict2[lista[i]][1] * self._dict2[lista[j]][1])
            self._mat_corr = matriz_corr.round(3)
            self._report_fit += "\nMatriz de correlação:\n\n" + self.matprint(
                self._mat_corr, ".3f") + "\n"
            self._report_fit += "Matriz de covariância:\n\n" + self.matprint(
                self._mat_cov, fmt = ".3e") + "\n"
            self._isvalid = True
        except TypeError:
            self._msg_handler.raise_error(
                "A função ajustada provavelmente não possui parâmetros para serem ajustados. Rever ajuste."
            )
            return None

    def __set_report_ODR(self, x):
        '''Constrói a string com os resultados.'''
        self._report_fit = ""
        self._report_fit += "\nAjuste: y = %s\n" % self._exp_model
        self._report_fit += "\nNGL  = %d" % (len(x) - len(self._par_var))
        self._report_fit += "\nChi² = %f\n\n" % self._result.sum_square
        self._report_fit += self.params_print()
        self._report_fit += "\n"
        self._mat_cov = self._result.cov_beta
        lista = list(self._params.keys())
        matriz_corr = np.zeros(
            (len(self._result.cov_beta), len(self._result.cov_beta)),
            dtype=float)
        z = range(len(matriz_corr))
        for i in z:
            for j in z:
                if self._dict[lista[i]][1] == 0 or self._dict[
                        lista[j]][1] == 0:
                    matriz_corr[i, j] = np.nan
                else:
                    matriz_corr[i, j] = self._result.cov_beta[i, j] / (
                        self._dict[lista[i]][1] * self._dict[lista[j]][1])
        self._mat_corr = matriz_corr.round(3)
        self._report_fit += "\nMatriz de correlação:\n\n" + self.matprint(
            self._mat_corr, ".3f") + "\n"
        self._report_fit += "Matriz de covariância:\n\n" + self.matprint(
            self._result.cov_beta, fmt = ".3e") + "\n"
        self._isvalid = True

    @property
    def coefficients(self):
        '''Retorna uma lista com os nomes dos coeficientes.'''
        return self._coef

    @property
    def xmin(self):
        return self.xmin_adj

    @xmin.setter
    def xmin(self, valor):
        self.xmin_adj = valor

    @property
    def xmax(self):
        return self.xmax_adj

    @xmax.setter
    def xmax(self, valor):
        self.xmax_adj = valor

    @property
    def data(self, *args):
        '''Retorna x, y, sx e sy.'''
        return self._data["x"], self._data["y"], self._data["sy"], self._data[
            "sx"]

    @property
    def residuo(self):
        '''Retorna os valores de y_i - f(x_i).'''
        return self._data["y"].to_numpy() - eval(f"self._model.eval({self._indVar}=self._data['x'].to_numpy())",
                                                 None, {"self" : self})

    @property
    def residuo_dummy(self):
        '''Retorna os valores de y_i - f(x_i).'''
        # self._coef = [i for i in self._model.param_names]
        # If there's no p0, everything is set to 1.0
        # pi = [0]*len(self._model.param_names)   # Inital values
        # if self._p0 is None:
        #     for i in range(len(self._model.param_names)):
        #         pi[i] = 1.0
        # else:
        #     for i in range(len(self._model.param_names)):
        #         try:
        #             pi[i] = float(self._p0[i])
        #         except:
        #             pi[i] = 1.0
        # paramss = Parameters()
        # for i in range(len(self._coef)):
        #     paramss.add(self._coef[i], pi[i])
        # print("rfs")
        # print("self._model.eval(%s = self._data['x'], params = self._params)"%self._indVar)
        # print(self._data["x"])
        # print(self._params)
        # print(self._model.eval(x = self._data['x'].to_numpy(), params = self._params))
        y = eval(
            "self._model.eval(%s = self._data['x'].to_numpy(), params = self._params)"
            % self._indVar, None, {"self": self})
        # print(y)
        # return self._data["y"].to_numpy() - eval("self._model.eval(%s = self._data['x'], params = self._params)"%self._indVar, None,
        # {"self": self})
        return self._data["y"].to_numpy() - y

    def get_predict(self, fig, x_min=None, x_max=None):
        '''Retorna a previsão do modelo.'''
        x_plot = np.linspace(x_min, x_max,
                             int(fig.get_size_inches()[0] * fig.dpi * 1.75))
        return x_plot, eval(
            "self._model.eval(%s = x_plot, params = self._params)" %
            self._indVar, None, {
                "x_plot": x_plot,
                "self": self
            })

    @property
    def inliers(self):
        '''Retorna os pontos usados no ajuste.'''
        return self._indices

    @property
    def outliers(self):
        '''Retorna os pontos não usados no ajuste.'''
        return np.array(list(set(np.arange(len(self._data))) - set(self._indices)))

    def get_predict_log(self, fig, x_min=None, x_max=None):
        '''Retorna a previsão do modelo.'''
        x_plot = np.logspace(np.log10(x_min), np.log10(x_max),
                             int(fig.get_size_inches()[0] * fig.dpi * 2.1))
        return x_plot, eval(
            "self._model.eval(%s = x_plot, params = self._params)" %
            self._indVar, None, {
                "x_plot": x_plot,
                "self": self
            })

    def predictInc(self, wsx, wsy: bool = False):
        if wsx == False and wsy == False and self._has_sx and self._has_sy:
            sy = np.zeros(len(self._data["x"]), dtype=float)
            for i, x in enumerate(self._data["x"]):
                x_var = np.array([
                    x + self._data["sx"].iloc[i], x - self._data["sx"].iloc[i]
                ])
                y_prd = eval(
                    "self._model.eval(%s = x, params = self._params)" %
                    self._indVar, None, {
                        "x": x,
                        "self": self
                    })
                y_var = eval(
                    "self._model.eval(%s = x_var, params = self._params)" %
                    self._indVar, None, {
                        "x_var": x_var,
                        "self": self
                    })
                sy[i] = np.abs(y_var - y_prd).mean()
                sy[i] = np.sqrt(self._data["sy"].iloc[i]**2 + sy[i]**2)
            return sy
        elif wsx == False and wsy == False and self._has_sy == False and self._has_sx:
            sy = np.zeros(len(self._data["x"]), dtype=float)
            for i, x in enumerate(self._data["x"]):
                x_var = np.array([
                    x + self._data["sx"].iloc[i], x - self._data["sx"].iloc[i]
                ])
                y_prd = eval(
                    "self._model.eval(%s = x, params = self._params)" %
                    self._indVar, None, {
                        "x": x,
                        "self": self
                    })
                y_var = eval(
                    "self._model.eval(%s = x_var, params = self._params)" %
                    self._indVar, None, {
                        "x_var": x_var,
                        "self": self
                    })
                sy[i] = np.abs(y_var - y_prd).mean()
        elif wsx == False and wsy == False and self._has_sy == False and self._has_sx == False:
            return np.zeros(len(self._data["x"]), dtype=float)
        elif wsx == False and wsy and self._has_sy and self._has_sx == False:
            return np.zeros(len(self._data["x"]), dtype=float)
        return self._data["sy"].to_numpy()

    def createDummyModel(self):
        try:
            self._model = ExpressionModel(self._exp_model +
                                          " + 0*%s" % self._indVar,
                                          independent_vars=[self._indVar])
        except ValueError:
            self._msg_handler.raise_error(
                "Expressão de ajuste escrita de forma errada. Rever função de ajuste."
            )
            return None
        except SyntaxError:
            self._msg_handler.raise_error(
                "Expressão de ajuste escrita de forma errada. Rever função de ajuste."
            )
            return None
        self._coef = [i for i in self._model.param_names]
        self.__make_parameters_lm()
        # if self._p0 is None:
        #     for i in range(len(self._coef)):
        #         self._params.add(self._coef[i], 1.)
        # else:
        #     coefs   = {c : [1, True] for c in self._coef}
        #     coefs_2 = {c : False for c in self._coef} # Para evitar de substituir atribuição de parâmetros
        #     for i in range(len(self._coef)):
        #         try:
        #             res = re.match("\s?(?:(.*[^\s])\s?=\s?(.*)|(@?\d*@?))", self._p0[i])
        #             if res.groups()[2] is None:
        #                     valor = res.groups()[1].replace("@", "")
        #                     var   = res.groups()[0]
        #                     if var in coefs:
        #                         coefs[var]   = [float(valor), not "@" in res.groups()[1]]
        #                         coefs_2[var] = True
        #             else:
        #                 if coefs_2[self._coef[i]] == False and self._coef[i] in coefs:
        #                     coefs[self._coef[i]]   = [float(res.groups()[2].replace("@", "")), not "@" in res.groups()[2]]
        #                     coefs_2[self._coef[i]] = True
        #         except:
        #             if coefs_2[self._coef[i]] == False:
        #                 coefs[self._coef[i]]   = [1, True]
        #                 coefs_2[self._coef[i]] = True
        #     for nome in coefs.keys():
        #         self._params.add(nome, coefs[nome][0], vary = coefs[nome][1])
        # coefs   = {c : [1, True] for c in self._coef}
        # coefs_2 = {c : False for c in self._coef} # Para evitar de substituir atribuição de parâmetros
        # for i in range(len(self._coef)):
        #     try:
        #         p_i = self._p0[i].split("=")
        #         if len(p_i) == 2:
        #             if "@" in p_i[1]:
        #                 p_i[1] = p_i[1].replace("@", "")
        #                 var = p_i[0].strip()
        #                 if var in coefs:
        #                     coefs[var] = [float(p_i[1].strip()), False]
        #                     coefs_2[var] = True
        #             else:
        #                 var = p_i[0].strip()
        #                 if var in coefs:
        #                     coefs[var]   = [float(p_i[1].strip()), True]
        #                     coefs_2[var] = True
        #         else:
        #             if "@" in self._p0[i]:
        #                 if coefs_2[self._coef[i]] == False:
        #                     coefs[self._coef[i]]   = [float(self._p0[i].replace("@", "")), False]
        #                     coefs_2[self._coef[i]] = True
        #             else:
        #                 if self._coef[i] in coefs and coefs_2[self._coef[i]] == False:
        #                     coefs[self._coef[i]]   = [float(self._p0[i]), True]
        #                     coefs_2[self._coef[i]] = True
        #     except:
        #         if coefs_2[self._coef[i]] == False:
        #             coefs[self._coef[i]]   = [1, True]
        #             coefs_2[self._coef[i]] = True
        # for nome in coefs.keys():
        #     self._params.add(nome, coefs[nome][0], vary = coefs[nome][1])
        self._indices = np.arange(len(self._data))
        self._isvalid = True

    def matprint(self, mat, fmt="f"):
        col_maxes = [
            max([len(("{:" + fmt + "}").format(x)) for x in col])
            for col in mat.T
        ]
        matrix = ""
        for x in mat:
            for i, y in enumerate(x):
                matrix += ("{:" + str(col_maxes[i]) + fmt +
                           "}").format(y) + "  "
            matrix += "\n"
        return matrix

    def params_print(self):
        df = pd.DataFrame(self._dict)
        df = df.transpose()
        df.columns = ["Valor", "|    Incerteza"]
        try:
            df.index = self._coef
        except:
            df.index = self._par_var
        return str(df)

    def params_print2(self, inc_considerada):
        df = pd.DataFrame(self._dict2).transpose()
        df.columns = ["Valor", "|    Incerteza"]
        df["|    Incerteza"] = df["|    Incerteza"] * inc_considerada
        df.index = self._par_var
        return str(df)

    def params_print3(self):
        df = pd.DataFrame(self._dict, columns=["Valor",
                                               "Incerteza"]).transpose()
        try:
            df.index = self._coef
        except:
            df.index = self._par_var
        return df

    @pyqtSlot(str, str, bool)
    def copyParamsClipboard(self, sep, decimal, header):
        '''Copy parameters to the clipboard.'''
        sep_decimal = {
            "Ponto": ".",
            "Vírgula": ",",
        }
        sep_columns = {
            "Tabulação": "\t",
            "Espaço": " ",
            ",": ",",
            "|": "|",
            ";": ";"
        }
        try:
            if header:
                pd.DataFrame(self._dict_param,
                             index=["Valor",
                                    "Incerteza"]).transpose().to_clipboard(
                                        sep=sep_columns[sep],
                                        decimal=sep_decimal[decimal])
            else:
                df = pd.DataFrame.from_dict(self._dict_param, orient="index")
                df.columns = [""] * len(df.columns)
                df.to_clipboard(sep=sep_columns[sep],
                                decimal=sep_decimal[decimal])

        except:
            self._msg_handler.raise_error(
                "Não foi possível copiar para a área de transferência.")

    @pyqtSlot(str, str)
    def copyCovarianceClipboard(self, sep, decimal):
        '''Copy parameters to the clipboard.'''
        sep_decimal = {
            "Ponto": ".",
            "Vírgula": ",",
        }
        sep_columns = {
            "Tabulação": "\t",
            "Espaço": " ",
            ",": ",",
            "|": "|",
            ";": ";"
        }
        try:
            pd.DataFrame(self._mat_cov).to_clipboard(
                sep=sep_columns[sep],
                decimal=sep_decimal[decimal],
                index=False,
                header=False)
        except:
            self._msg_handler.raise_error(
                "Não foi possível copiar para a área de transferência.")

    @pyqtSlot(str, str)
    def copyCorrelationClipboard(self, sep, decimal):
        '''Copy parameters to the clipboard.'''
        sep_decimal = {
            "Ponto": ".",
            "Vírgula": ",",
        }
        sep_columns = {
            "Tabulação": "\t",
            "Espaço": " ",
            ",": ",",
            "|": "|",
            ";": ";"
        }
        try:
            pd.DataFrame(self._mat_corr).to_clipboard(
                sep=sep_columns[sep],
                decimal=sep_decimal[decimal],
                index=False,
                header=False)
        except:
            self._msg_handler.raise_error(
                "Não foi possível copiar para a área de transferência.")

    def reset(self):
        self._data = None
        self._data_json = None
        self._exp_model = ""
        self._model = None
        self._report_fit = ""
        self._result = None
        self._coef = []
        self._params = Parameters()
        self._dict = {}
        self._dict2 = {}
        self._p0 = None
        self.xmin_adj = 0.
        self.xmax_adj = 0.
        self._mode = 0
        self._has_data = False
        self._isvalid = False
        self._has_sx = True
        self._has_sy = True
        self._indices = []



# -*- coding: utf-8 -*-
"""
MIT License

Copyright (c) 2021 Leonardo Eiji Tamayose, Guilherme Ferrari Fortino

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from PyQt5.QtCore import QObject, QJsonValue, QUrl, pyqtSignal, pyqtSlot
from src.Calculators import interpreter_calculator, plot
from matplotlib import colors
import numpy as np
import pandas as pd
import json
import platform

class SinglePlot(QObject):
    '''Class that controls the single-plot page'''

    # Signal to write infos
    write_calculator       = pyqtSignal(str, arguments='expr')
    fill_plot_page_signal = pyqtSignal(QJsonValue, arguments='props')
    plot_signal           = pyqtSignal()

    def __init__(self, canvas, model, messageHandler):
        super().__init__()
        self.canvas = canvas
        self.model  = model
        self.path   = ''
        self.msg    = messageHandler

        # Default properties for the singlePlot page
        self.props = {
            'id': '',
            'dataProps': {
                'marker_color'    : '#000',
                'marker_size'     : 3,
                'marker'          : 'o',
                'curve_color'     : '#000',
                'curve_thickness' : 3,
                'curve_style'     : '-',
            },
            'canvasProps': {
                'xaxis'     : '',
                'yaxis'     : '',
                'title'     : '',
                'log_x'     : False,
                'log_y'     : False,
                'legend'    : False,
                'grid'      : False,
                'residuals' : False,
                'xmin'      : '',
                'xmax'      : '',
                'xdiv'      : '',
                'ymin'      : '',
                'ymax'      : '',
                'ydiv'      : '',
                'resmin'    : '',
                'resmax'    : '',
            },
            'fitProps': {
                'expr'       : '',
                'p0'         : '',
                'wsx'        : True,
                'wsy'        : True,
                'xmin'       : '',
                'xmax'       : '',
                'parameters' : {},
                'adjust'     : True
            },
            'data': []
        }

    @pyqtSlot(QJsonValue)
    def get_plot_data(self, plotData):
        self.model.reset()
        plotData    = plotData.toVariant()
        canvasProps = plotData['canvasProps']
        dataProps   = plotData['dataProps']
        fitProps    = plotData['fitProps']

        # Loading data from the table
        self.model.loadDataTable(plotData['data'])

        # Getting function to fit
        # Anti-dummies system
        fitProps['expr'] = fitProps['expr'].replace('^', '**').replace('arctan', 'atan').replace('arcsin', 'asin')
        fitProps['expr'] = fitProps['expr'].replace('arccos', 'acos').replace('sen', 'sin').replace('raiz', 'sqrt')
        expIndVar = fitProps['expr'].split(";")
        # Setting expression
        if len(expIndVar) == 2:
            if self.model._exp_model != expIndVar[0]:
                self.model.set_expression(expIndVar[0].strip(), expIndVar[1].strip())
        elif len(expIndVar) == 1:
            if self.model._exp_model != expIndVar[0]:
                self.model.set_expression(expIndVar[0])

        # Getting initial parameters
        if fitProps['p0'].strip() != '':
            p0 = fitProps['p0']
            # p0 = p0.replace(';', ',')
            p0 = p0.replace('/', ',')
            self.model.set_p0(p0)
        
        self.model.xmin = self.make_float(fitProps['xmin'], value = -np.inf)
        self.model.xmax = self.make_float(fitProps['xmax'], value = np.inf)

        if self.model.xmin >= self.model.xmax:
            self.msg.raise_error("Intervalo de ajuste inválido. Rever intervalo de ajuste.")
            return None

        # Setting style of the plot
        self.plot(self.model, canvasProps, fitProps, dataProps)

    def plot_in_out(self, x_i, y_i, x_o, y_o, kargs_errorbar, y_ri = None, y_ro = None, sy_i = None, sy_o = None, sx_i = None, sx_o = None, ssy_i = None, ssy_o = None):
        '''Macro para o plot dos ditos inliers e outliers de um ajuste.'''
        self.canvas.plot_error_bar(x_i, y_i, kargs_errorbar,
            alpha = 1, sy = sy_i, sx = sx_i, y_r = y_ri, ssy = ssy_i)
        if self.canvas.user_color_outliers is not None:
            kargs_errorbar["ecolor"] = self.canvas.user_color_outliers
            kargs_errorbar["c"] = self.canvas.user_color_outliers
        self.canvas.plot_error_bar(x_o, y_o, kargs_errorbar,
            alpha = self.canvas.user_alpha_outliers, sy = sy_o,
            y_r = y_ro, sx = sx_o, ssy = ssy_o)

    
    def plot(self, model, canvas_props, fit_props, data_props):
        self.canvas.set_tight_layout()
        sigma_x = not not fit_props["wsx"]
        sigma_y = not not fit_props["wsy"]
        grid = not not canvas_props["grid"]
        log_x = not not canvas_props["log_x"]
        log_y = not not canvas_props["log_y"]
        legend = not not canvas_props["legend"]
        residuals = not not canvas_props["residuals"]
        xmin = canvas_props["xmin"]
        xmax = canvas_props["xmax"]
        xdiv = canvas_props["xdiv"]
        ymin = canvas_props["ymin"]
        ymax = canvas_props["ymax"]
        ydiv = canvas_props["ydiv"]
        resmin = canvas_props["resmin"]
        resmax = canvas_props["resmax"]
        partial_titles = canvas_props["title"].split(";")
        symbol_color = data_props["marker_color"]
        symbol_size = data_props["marker_size"]
        symbol = data_props["marker"]
        curve_color = data_props["curve_color"]
        curve_thickness = data_props["curve_thickness"]
        curve_style = data_props["curve_style"]
        px, py, y_r = None, None, None
        self.canvas.grid = grid
        axis_titles = []
        if len(partial_titles) == 1:
            axis_titles = [
                canvas_props["title"].strip(),
                canvas_props["xaxis"].strip(),
                canvas_props["yaxis"].strip(),
                "",
            ]
        else:
            axis_titles = [
                partial_titles[0].strip(),
                canvas_props["xaxis"].strip(),
                canvas_props["yaxis"].strip(),
                partial_titles[1].strip(),
            ]
        if model._has_data:

            # Fitting expression to data, if there's any expression
            if fit_props["adjust"]:
                if model._exp_model != "":
                    model.fit(wsx=not sigma_x, wsy=not sigma_y)
                else:
                    model._isvalid = False
            else:
                if model._exp_model != "":
                    model.createDummyModel()
                else:
                    model._isvalid = False

            kargs_errorbar = { "ecolor" : symbol_color,
                               "capsize" : 0,
                               "elinewidth" : 1,
                               "ms" : symbol_size,
                               "marker" : symbol,
                               "c" : symbol_color,
                               "ls" : "none"}

            # Plotting if the model is valid
            if model._isvalid:
                # Clearing the current plot
                self.canvas.clear_axis()

                # Getting data
                x, y, sy, sx = model.data
                inliers, outliers = model.inliers, model.outliers
                x_i, y_i, sy_i, sx_i = x.iloc[inliers], y.iloc[inliers], sy.iloc[inliers], sx.iloc[inliers]
                x_o, y_o, sy_o, sx_o = np.array([]), np.array([]), np.array([]), np.array([])
                if len(outliers) > 0:
                    x_o, y_o, sy_o, sx_o = x.iloc[outliers], y.iloc[outliers], sy.iloc[outliers], sx.iloc[outliers]
                # alphas = np.array([list(colors.to_rgba(symbol_color))]*len(x))
                # alphas[model.indices.astype(int), 3] = self.canvas.user_alpha_outliers
                # kargs_errorbar["c"] = alphas
                # kargs_errorbar["ecolor"] = alphas
                y_r = None
                if fit_props["adjust"]:
                    y_r = model.residuo
                    y_ri = y_r[inliers]
                    y_ro = np.array([])
                    if len(outliers) > 0:
                        y_ro = y_r[outliers]
                else:
                    y_r = model.residuo_dummy
                    y_ri, y_ro = np.copy(y_r), np.array([])
                if residuals:
                    self.canvas.switch_axes(hide_axes2=False)
                    if sigma_x and sigma_y:  # Caso considerar as duas incertezas
                        ssy = model.predictInc(not sigma_x)
                        ssy_i = ssy[inliers]
                        ssy_o = np.array([])
                        if len(outliers) > 0:
                            ssy_o = ssy[outliers]
                        self.plot_in_out(x_i, y_i, x_o, y_o,
                         kargs_errorbar, y_ri = y_ri, y_ro = y_ro, sy_i = sy_i,
                         sy_o = sy_o, sx_i = sx_i, sx_o = sx_o,
                         ssy_i = ssy_i, ssy_o = ssy_o)
                    elif (sigma_x is False
                          and sigma_y is False):  # Caso desconsiderar as duas
                        self.plot_in_out(x_i, y_i, x_o, y_o, kargs_errorbar,
                        y_ri = y_ri, y_ro = y_ro)
                    elif sigma_x is False and sigma_y is True:  # Caso considerar só sy
                        ssy = model.predictInc(not sigma_x)
                        ssy_i = ssy[inliers]
                        ssy_o = np.array([])
                        if len(outliers) > 0:
                            ssy_o = ssy[outliers]
                        self.plot_in_out(x_i, y_i, x_o, y_o,
                         kargs_errorbar, y_ri = y_ri, y_ro = y_ro, sy_i = sy_i,
                         sy_o = sy_o, ssy_i = ssy_i, ssy_o = ssy_o)
                    else:  # Caso considerar só sx
                        ssy = model.predictInc(not sigma_x, not sigma_y)
                        ssy_i = ssy[inliers]
                        ssy_o = np.array([])
                        if len(outliers) > 0:
                            ssy_o = ssy[outliers]
                        self.plot_in_out(x_i, y_i, x_o, y_o,
                         kargs_errorbar, y_ri = y_ri, y_ro = y_ro,
                         ssy_i = ssy_i, ssy_o = ssy_o)
                    self.canvas.set_axes_props_with_axes_2(
                        xmin,
                        xmax,
                        xdiv,
                        ymin,
                        ymax,
                        ydiv,
                        resmin,
                        resmax,
                        grid,
                        log_x,
                        log_y,
                    )
                    left, right = self.canvas.axes1.get_xlim()
                    px, py = 0.0, 0.0
                    if log_x:
                        px, py = model.get_predict_log(self.canvas.axes1.figure, left,
                                                       right)
                    else:
                        px, py = model.get_predict(self.canvas.axes1.figure, left,
                                                   right)

                    self.canvas.axes2.axline((0., 0.), slope = 0., color = curve_color,
                    alpha = 0.65, zorder = 0)

                    # Making Plots
                    (line_func, ) = self.canvas.axes1.plot(
                        px,
                        py,
                        lw=curve_thickness,
                        color=curve_color,
                        ls=curve_style,
                        label=f"${model._exp_model}$",
                    )

                    # Setting titles
                    self.canvas.axes1.set_title(axis_titles[0],
                                             fontsize=self.canvas.font_sizes["titulo"])
                    self.canvas.axes2.set_xlabel(xlabel=axis_titles[1],
                                             fontsize=self.canvas.font_sizes["eixo_x"])
                    self.canvas.axes1.set_ylabel(ylabel=axis_titles[2],
                                             fontsize=self.canvas.font_sizes["eixo_y"])
                    self.canvas.axes2.set_ylabel(ylabel=axis_titles[3],
                                             fontsize=self.canvas.font_sizes["residuos"])
                    if legend:
                        self.canvas.axes1.legend(frameon=False,
                                            fontsize = self.canvas.font_sizes["legenda"],
                                            loc = self.canvas.legend_loc)

                    def update(evt):
                        left, right = self.canvas.axes1.get_xlim()
                        ppx, ppy = model.get_predict(self.canvas.axes1.figure, left,
                                                     right)
                        line_func.set_data(ppx, ppy)
                        self.canvas.axes1.figure.canvas.draw_idle()

                    if log_x:

                        def update(evt):
                            left, right = self.canvas.axes1.get_xlim()
                            ppx, ppy = model.get_predict_log(
                                self.canvas.axes1.figure, left, right)
                            line_func.set_data(ppx, ppy)
                            self.canvas.axes1.figure.canvas.draw_idle()

                    self.canvas.axes1.remove_callback(self.canvas.oid)
                    self.canvas.axes1.figure.canvas.mpl_disconnect(self.canvas.cid)
                    self.canvas.oid = self.canvas.axes1.callbacks.connect(
                        "xlim_changed", update)
                    self.canvas.cid = self.canvas.axes1.figure.canvas.mpl_connect(
                        "resize_event", update)
                else:
                    self.canvas.clear_axis()
                    self.canvas.switch_axes(hide_axes2=True)

                    # Making Plots
                    if sigma_x and sigma_y:  # Caso considerar as duas incertezas
                        self.plot_in_out(x_i, y_i, x_o, y_o,
                         kargs_errorbar, sy_i = sy_i, sy_o = sy_o)
                    elif (sigma_x is False
                          and sigma_y is False):  # Caso desconsiderar as duas
                        self.plot_in_out(x_i, y_i, x_o, y_o,
                         kargs_errorbar)
                    elif sigma_x is False and sigma_y is True:  # Caso considerar só sy
                        self.plot_in_out(x_i, y_i, x_o, y_o,
                         kargs_errorbar, sy_i = sy_i, sy_o = sy_o)
                    else:  # Caso considerar só sx
                        self.plot_in_out(x_i, y_i, x_o, y_o,
                         kargs_errorbar, sx_i = sx_i, sx_o = sx_o)

                    self.canvas.set_axes_props_without_axes_2(xmin, xmax, xdiv, ymin,
                                                       ymax, ydiv, grid, log_x,
                                                       log_y)
                    left, right = self.canvas.axes1.get_xlim()
                    px, py = 0.0, 0.0
                    if log_x:
                        px, py = model.get_predict_log(self.canvas.axes1.figure, left,
                                                       right)
                    else:
                        px, py = model.get_predict(self.canvas.axes1.figure, left,
                                                   right)

                    (line_func, ) = self.canvas.axes1.plot(
                        px,
                        py,
                        lw=curve_thickness,
                        color=curve_color,
                        ls=curve_style,
                        label=f"${model._exp_model}$",
                        picker=True,
                    )
                    if legend:
                        self.canvas.axes1.legend(frameon=False,
                                            fontsize = self.canvas.font_sizes["legenda"],
                                            loc = self.canvas.legend_loc)

                    # Setting titles
                    self.canvas.axes1.set_title(str(axis_titles[0]), 
                                            fontsize=self.canvas.font_sizes["titulo"])
                    self.canvas.axes1.set_xlabel(xlabel=str(axis_titles[1]),
                                            fontsize=self.canvas.font_sizes["eixo_x"])
                    self.canvas.axes1.set_ylabel(ylabel=str(axis_titles[2]),
                                            fontsize=self.canvas.font_sizes["eixo_y"])

                    # One piece
                    def update(evt):
                        left, right = self.canvas.axes1.get_xlim()
                        ppx, ppy = model.get_predict(self.canvas.axes1.figure, left,
                                                     right)
                        line_func.set_data(ppx, ppy)
                        self.canvas.axes1.figure.canvas.draw_idle()

                    if log_x:

                        def update(evt):
                            left, right = self.canvas.axes1.get_xlim()
                            ppx, ppy = model.get_predict_log(
                                self.canvas.axes1.figure, left, right)
                            line_func.set_data(ppx, ppy)
                            self.canvas.axes1.figure.canvas.draw_idle()

                    self.canvas.axes1.remove_callback(self.canvas.oid)
                    self.canvas.axes1.figure.canvas.mpl_disconnect(self.canvas.cid)
                    self.canvas.oid = self.canvas.axes1.callbacks.connect(
                        "xlim_changed", update)
                    self.canvas.cid = self.canvas.figure.canvas.mpl_connect(
                        "resize_event", update)

            else:
                self.canvas.clear_axis()
                self.canvas.switch_axes(hide_axes2=True)

                x, y, sy, sx = model.data

                # Making Plots
                if sigma_x and sigma_y:  # Caso considerar as duas incertezas
                    self.canvas.plot_error_bar(x, y, kargs_errorbar,
                        alpha = 1, sx = sx, sy = sy)
                elif (sigma_x is False
                      and sigma_y is False):  # Caso desconsiderar as duas
                    self.canvas.plot_error_bar(x, y, kargs_errorbar,
                        alpha = 1)
                elif sigma_x is False and sigma_y is True:  # Caso considerar só sy
                    self.canvas.plot_error_bar(x, y, kargs_errorbar,
                        alpha = 1, sy = sy)
                else:  # Caso considerar só sx
                    self.canvas.plot_error_bar(x, y, kargs_errorbar,
                        alpha = 1, sx = sx)

                # Setting titles
                self.canvas.axes1.set_title(str(axis_titles[0]), 
                                            fontsize=self.canvas.font_sizes["titulo"])
                self.canvas.axes1.set_xlabel(xlabel=str(axis_titles[1]),
                                            fontsize=self.canvas.font_sizes["eixo_x"])
                self.canvas.axes1.set_ylabel(ylabel=str(axis_titles[2]),
                                            fontsize=self.canvas.font_sizes["eixo_y"])
                self.canvas.set_axes_props_without_axes_2(xmin, xmax, xdiv, ymin,
                                                   ymax, ydiv, grid, log_x,
                                                   log_y)

        # Reseting parameters
        model.isvalid = False
        self.canvas.canvas.draw_idle()

    def fill_plot_page(self, props=None):
        # If no properties passed, emit the default values
        if props is None:
            self.fill_plot_page_signal.emit(QJsonValue.fromVariant(self.props))
        else:
            self.fill_plot_page_signal.emit(QJsonValue.fromVariant(props))
        
    @pyqtSlot()
    def new(self):
        # Reseting canvas and model
        self.model.reset()
        # self.canvas.reset()
        self.canvas.clear_axis()
        self.canvas.switch_axes(True)

        # Fill singlePlot page with default values
        self.fill_plot_page()

        # Reseting path
        self.path = ''

    @pyqtSlot(str)
    def load(self, path):
        # Reseting frontend
        self.new()

        # Getting path
        self.path = QUrl(path).toLocalFile()

        # Getting props
        with open(self.path, encoding='utf-8') as file:
            props = json.load(file)

        if "key" in props:
            if props["key"] != "2-b":
                self.msg.raise_warn("O carregamento de arquivos antigos está limitado à uma versão anterior. Adaptação feita automaticamente.")
            if props["key"].split('-')[-1] == 'multiplot':
                self.msg.raise_error("O projeto carregado pertence ao multiplot, esse arquivo é incompatível.")
                return 0
            elif props["key"].split('-')[-1] == 'hist':
                self.msg.raise_error("O projeto carregado pertence ao histograma, esse arquivo é incompatível.")
                return 0
            # Loading data from the project
            self.model.load_data(df_array=props['data'])
        else:
            try:
                self.msg.raise_warn("O carregamento de arquivos antigos está limitado à uma versão anterior. Adaptação feita automaticamente.")
                props = self.load_old_json(props)
            except:
                self.msg.raise_error("O arquivo carregado é incompatível com o ATUS.")
                return 0
            self.model.load_data(df=props['data'])

        self.fill_plot_page(props)

    def load_old_json(self, props):
        props_tmp = self.props.copy()

        # Shaping old json into the new one
        props_tmp['id']                           = props['projectName']
        props_tmp['dataProps']['marker_color']    = props['symbol_color']
        props_tmp['dataProps']['marker_size']     = props['symbol_size']
        props_tmp['dataProps']['marker']          = props['symbol']
        props_tmp['dataProps']['curve_color']     = props['curve_color']
        props_tmp['dataProps']['curve_thickness'] = props['curve_thickness']
        props_tmp['dataProps']['curve_style']     = props['curve_style']
        props_tmp['canvasProps']['xaxis']         = props['xaxis']
        props_tmp['canvasProps']['yaxis']         = props['yaxis']
        props_tmp['canvasProps']['title']         = props['title']
        props_tmp['canvasProps']['log_x']         = props['log_x']
        props_tmp['canvasProps']['log_y']         = props['log_y']
        props_tmp['canvasProps']['legend']        = props['legend']
        props_tmp['canvasProps']['grid']          = props['grid']
        props_tmp['canvasProps']['residuals']     = props['residuals']
        props_tmp['canvasProps']['xmin']          = props['xmin']
        props_tmp['canvasProps']['xmax']          = props['xmax']
        props_tmp['canvasProps']['xdiv']          = props['xdiv']
        props_tmp['canvasProps']['ymin']          = props['ymin']
        props_tmp['canvasProps']['ymax']          = props['ymax']
        props_tmp['canvasProps']['ydiv']          = props['ydiv']
        props_tmp['canvasProps']['resmin']        = props['resmin']
        props_tmp['canvasProps']['resmax']        = props['resmax']
        props_tmp['fitProps']['expr']             = props['expr']
        props_tmp['fitProps']['p0']               = props['p0']
        props_tmp['fitProps']['wsx']              = props['wsx']
        props_tmp['fitProps']['wsy']              = props['wsy']
        props_tmp['fitProps']['parameters']       = props['parameters']
        props_tmp['data']                         = pd.read_json(props['data'], dtype=str)

        return props_tmp

    def make_float(self, var, value = 0.):
        try:
            return float(var)
        except:
            return value
            
    def make_int(self, var, value = 0):
        try:
            return int(var)
        except:
            return value

    @pyqtSlot(QJsonValue, result=int)
    def save(self, props):
        # If there's no path for saving, save_as()
        if self.path == '':
            return 1

        # Getting properties
        props                           = props.toVariant()
        props["fitProps"]["parameters"] = self.model._params.valuesdict()

        if platform.system() == "Linux":
            if self.path[-5:] == ".json":
                with open(self.path, 'w', encoding='utf-8') as file:
                    json.dump(props, file, ensure_ascii=False, indent=4)
            else: 
                with open(self.path + ".json", 'w', encoding='utf-8') as file:
                    json.dump(props, file, ensure_ascii=False, indent=4)
        else:
            with open(self.path, 'w', encoding='utf-8') as file:
                json.dump(props, file, ensure_ascii=False, indent=4)

        return 0
    
    @pyqtSlot(str, QJsonValue)
    def save_as(self, path, props):
        # Getting path
        self.path = QUrl(path).toLocalFile()

        # Getting properties
        props = props.toVariant()
        props['fitProps']['parameters'] = self.model._params.valuesdict()

        if platform.system() == "Linux":
            if self.path[-5:] == ".json":
                with open(self.path, 'w', encoding='utf-8') as file:
                    json.dump(props, file, ensure_ascii=False, indent=4)
            else: 
                with open(self.path + ".json", 'w', encoding='utf-8') as file:
                    json.dump(props, file, ensure_ascii=False, indent=4)
        else:
            with open(self.path, 'w', encoding='utf-8') as file:
                json.dump(props, file, ensure_ascii=False, indent=4)

    @pyqtSlot(str, str, str, str, str, str)
    def calculator(self, function, opt1, nc, ngl, mean, std):
        functionDict = {
            'Chi²':0,
            'Chi² Reduzido':1,
            'Gaussiana':2,
            'Student':3
        }
        methodDict = {
            'Simétrico de Dois Lados':0,
            'Apenas Limite Inferior':1,
            'Apenas Limite Superior':2
        }   
        try:
            nc = nc.replace(',', '.')
            nc = float(nc)
            if nc == 0 or nc >= 1:
                self.msg.raise_error("Nível de confiança deve ser sempre maior que zero e menor que 1. Rever nível de confiança.")
                return None
        except:
            pass
        try:
            ngl = ngl.replace(',', '.')
            ngl = float(ngl)
        except:
            pass
        try:
            mean = mean.replace(',', '.')
            mean = float(mean)
        except:
            pass
        try:
            std = std.replace(',', '.')
            std = float(std)
            if std <= 0:
                self.msg.raise_error("Desvio padrão deve ser sempre maior que zero. Rever desvio padrão.")
                return None
        except:
            pass
        
        s, x, y, x_area, y_area, title, xlabel, ylabel = interpreter_calculator(functionDict[function], methodDict[opt1], nc, ngl, mean, std)
        plot(self.canvas, x, y, x_area, y_area, title, xlabel, ylabel)
        self.write_calculator.emit(s)

    @pyqtSlot(QJsonValue)
    def export_data_clipboard(self, data):
        df = pd.DataFrame.from_records(data.toVariant())

        if df.empty:
            self.msg.raise_warn("Nenhum dado para exportar.")
            return

        df.columns = ['x', 'y', 'sy', 'sx', 'bool']
        del df['bool']
        df.to_clipboard(index=False)

        self.msg.raise_success("Dados copiados para área de transferência.")



# -*- coding: utf-8 -*-
"""
MIT License

Copyright (c) 2021 Leonardo Eiji Tamayose, Guilherme Ferrari Fortino

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import os
import sys

import matplotlib.pyplot as plt
from matplotlib_backend_qtquick.backend_qtquickagg import FigureCanvasQtQuickAgg
from PyQt5.QtCore import QCoreApplication, QObject, Qt, QUrl, QThread
from PyQt5.QtGui import QGuiApplication, QIcon
from PyQt5.QtQml import QQmlApplicationEngine, qmlRegisterType
from PyQt5.QtWidgets import QApplication

sys.path.append(os.path.dirname(os.path.abspath(__file__)))
from src.GlobalManager import GlobalManager
from src.GoogleDriveAPI import GDrive
from src.Histogram import Histogram
from src.MatPlotLib import Canvas
from src.MessageHandler import MessageHandler
from src.Model import Model
from src.MultiPlot import Multiplot
from src.Plot import SinglePlot
from src.UpdateChecker import UpdateChecker
from src.PyLatex import PyLatex

plt.rcParams["ytick.minor.visible"] = False
plt.rcParams["xtick.minor.visible"] = False
plt.rcParams["figure.subplot.left"] = 0.1
plt.rcParams["figure.subplot.right"] = 0.95
plt.rcParams["figure.subplot.bottom"] = 0.12
plt.rcParams["figure.subplot.top"] = 0.92
plt.rcParams["figure.subplot.hspace"] = 0.0


def main(pip: bool = True):
    if pip:
        print("Initializing Analysis Tool for Undergrad Students...")

    # Matplotlib stuff
    qmlRegisterType(FigureCanvasQtQuickAgg, "Canvas", 1, 0, "FigureCanvas")
    QCoreApplication.setAttribute(Qt.AA_EnableHighDpiScaling)

    # Setting up app
    app = QApplication(sys.argv)
    app.setOrganizationName("High Elo Devs")
    app.setOrganizationDomain(
        "https://github.com/leoeiji/Analysis-Tool-for-Undergrad-Students---ATUS"
    )
    app.setApplicationName("Analysis Tool for Undergrad Students")
    app.setWindowIcon(
        QIcon(
            os.path.join(
                os.path.dirname(__file__), "images/main_icon/ATUS_icon.png"
            )
        )
    )
    engine = QQmlApplicationEngine()

    messageHandler = MessageHandler()
    canvas = Canvas(messageHandler)
    model = Model(messageHandler)
    singlePlot = SinglePlot(canvas, model, messageHandler)
    multiPlot = Multiplot(canvas, messageHandler)
    updater = UpdateChecker(pip)
    histogram = Histogram(canvas, messageHandler)
    gdrive = GDrive(messageHandler)
    globalManager = GlobalManager()
    pylatex = PyLatex()

    thread = QThread()
    thread.start(5)
    pylatex.moveToThread(thread)

    # Creating 'link' between front-end and back-end
    context = engine.rootContext()
    context.setContextProperty("singlePlot", singlePlot)
    context.setContextProperty("multiPlot", multiPlot)
    context.setContextProperty("canvas", canvas)
    context.setContextProperty("model", model)
    context.setContextProperty("updater", updater)
    context.setContextProperty("messageHandler", messageHandler)
    context.setContextProperty("hist", histogram)
    context.setContextProperty("gdrive", gdrive)
    context.setContextProperty("globalManager", globalManager)
    context.setContextProperty("pylatex", pylatex)

    # Loading canvas window
    engine.load(
        QUrl.fromLocalFile(
            os.path.join(
                os.path.dirname(__file__), "qml/controls/CanvasWindow.qml"
            )
        )
    )
    context.setContextProperty("canvasWindow", engine.rootObjects()[0])

    # Loading main window
    engine.load(
        QUrl.fromLocalFile(
            os.path.join(os.path.dirname(__file__), "qml/main_window.qml")
        )
    )

    # Updating canvasPlot with the plot
    canvas.update_with_canvas(
        engine.rootObjects()[1].findChild(QObject, "canvasPlot")
    )

    # Stopping program if PyQt fails loading the file
    if not engine.rootObjects():
        sys.exit(-1)

    # Starting program
    sys.exit(app.exec_())


if __name__ == "__main__":
    main(False)



# -*- coding: utf-8 -*-
"""
MIT License

Copyright (c) 2021 Leonardo Eiji Tamayose, Guilherme Ferrari Fortino

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import json
import numpy as np
import platform
from PyQt5.QtCore import QObject, QJsonValue, QUrl, pyqtSignal, pyqtSlot
from .Model_multiplot import MultiModel

class Multiplot(QObject):
    """Backend for multiplot page"""

    setData          = pyqtSignal(QJsonValue, arguments='data')
    removeRow        = pyqtSignal(int, arguments='row')
    fillPageSignal   = pyqtSignal(QJsonValue, arguments='props')
    addRow           = pyqtSignal(QJsonValue, arguments='rowData')

    def __init__(self, displayBridge, messageHandler):
        super().__init__()
        self.path          = ''
        self.displayBridge = displayBridge
        self.msg           = messageHandler
        self.Multi_Model   = None
        self.grid          = 0.
        self.xmin          = 0.
        self.xmax          = 0.
        self.xdiv          = 0.
        self.ymin          = 0.
        self.ymax          = 0.
        self.ydiv          = 0.
        self.logy          = 0.
        self.logx          = 0.
        self.title         = ''
        self.xaxis         = ''
        self.yaxis         = ''

        self.defaultProps  = {
            'id': '',
            'rowsData': [],
            'canvasProps': {
                'title': '',
                'xaxis': '',
                'yaxis': '',
                'xmin': '',
                'xmax': '',
                'xdiv': '',
                'ymin': '',
                'ymax': '',
                'ydiv': '',
                'logx': False,
                'logy': False,
                'grid': False,
            }
        }

    def fillPage(self, props):
        props = QJsonValue.fromVariant(props)
        self.fillPageSignal.emit(props)

    @pyqtSlot()
    def new(self):
        # Reseting path
        self.path = ''

        # Reseting frontend
        self.displayBridge.clear_axis()
        self.displayBridge.switch_axes(True)
        self.fillPage(self.defaultProps)

    @pyqtSlot(str)
    def load(self, path):
        curveStyles = {
            '-': 0,
            '--': 1,
            '-.':2
        }
        # Reseting frontend
        self.new()

        # Setting path
        self.path = QUrl(path).toLocalFile()

        # Getting props
        with open(self.path, encoding='utf-8') as file:
            try:
                props = json.load(file)
            except:
                self.msg.raise_error("O arquivo carregado é incompatível.")

        if "key" in props:
            # Loading data from the table
            key = props["key"].split('-')
            if key[0] != "2":
                self.msg.raise_warn("O carregamento de arquivos antigos está limitado à uma versão anterior.")
                return 0
            if key[-1] != 'multiplot':
                self.msg.raise_error("O arquivo carregado é incompatível ou está desatualizado.")
                return 0
        else:
            self.msg.raise_error("O arquivo carregado é incompatível ou está desatualizado.")
            return 0

        for row, rowData in enumerate(props['rowsData'], start=0):
            prop = QJsonValue.fromVariant({
                'row': row,
                'data': rowData['df'],
                'params': rowData['params'],
                'fileName': 'Carregado do projeto',
                'projectName': rowData['label'],
                'expr': rowData['expr'],
                'p0': rowData['p0'],
                'symbolColor': rowData["markerColor"],
                'curve': curveStyles[rowData['curve']]
            })
            self.addRow.emit(prop)

        # Removing rowData from props
        del props['rowsData']

        # Filling page
        self.fillPageSignal.emit(QJsonValue.fromVariant(props))

    @pyqtSlot(QJsonValue, result=int)
    def save(self, props):
        # If there's no path for saving, save_as()
        if self.path == '':
            return 1

        # Getting properties
        props = props.toVariant()

        if platform.system() == "Linux":
            if self.path[-5:] == ".json":
                with open(self.path, 'w', encoding='utf-8') as file:
                    json.dump(props, file, ensure_ascii=False, indent=4)
            else: 
                with open(self.path + ".json", 'w', encoding='utf-8') as file:
                    json.dump(props, file, ensure_ascii=False, indent=4)
        else:
            with open(self.path, 'w', encoding='utf-8') as file:
                json.dump(props, file, ensure_ascii=False, indent=4)

        return 0
    
    @pyqtSlot(str, QJsonValue)
    def save_as(self, path, props):
        # Getting path
        self.path = QUrl(path).toLocalFile()

        # Getting properties
        props = props.toVariant()

        if platform.system() == "Linux":
            if self.path[-5:] == ".json":
                with open(self.path, 'w', encoding='utf-8') as file:
                    json.dump(props, file, ensure_ascii=False, indent=4)
            else: 
                with open(self.path + ".json", 'w', encoding='utf-8') as file:
                    json.dump(props, file, ensure_ascii=False, indent=4)
        else:
            with open(self.path, 'w', encoding='utf-8') as file:
                json.dump(props, file, ensure_ascii=False, indent=4)

    @pyqtSlot(str, int)
    def loadData(self, fileUrl, row):
        curveStyles = {
            '-': 0,
            '--': 1,
            '-.':2
        }
        # Opening json file
        with open(QUrl(fileUrl).toLocalFile(), encoding='utf-8') as file:
            data = json.load(file)

        try:
            # Getting function to fit
            # Anti-dummies system
            expr = data["fitProps"]['expr']
            if expr != '':
                expr = expr.replace('^', '**')
                expr = expr.replace('arctan', 'atan')
                expr = expr.replace('arcsin', 'asin')
                expr = expr.replace('arccos', 'acos')
                expr = expr.replace('sen', 'sin')
                
            self.setData.emit(QJsonValue.fromVariant({
                'row': row,
                'data': data['data'],
                'params': data["fitProps"]['parameters'],
                'fileName': QUrl(fileUrl).toLocalFile().split('/')[-1],
                'projectName': data['id'],
                'expr': expr,
                'p0': data["fitProps"]['p0'],
                'symbolColor': data["dataProps"]["marker_color"],
                'curve': curveStyles[data["dataProps"]['curve_style']]
            }))
        except:
            self.msg.raise_error("Erro ao carregar arquivo. Verificar arquivo de entrada.")
            self.removeRow.emit(row)

    @pyqtSlot(QJsonValue)
    def getData(self, data):
        '''Get data from frontend and make a plot'''

        def mk_float(s):
            '''Make a float from the string'''
            s = s.strip()
            return np.float64(s) if s else 0
        def mk_int(s):
            '''Make a float from the string'''
            s = s.strip()
            return np.int64(s) if s else 0

        dados              = data.toVariant()
        graph_options      = dados['canvasProps']
        projetos           = dados['rowsData']
        self.Multi_Model   = MultiModel(graph_options, projetos)
        self.grid          = graph_options['grid']
        self.logx          = graph_options['logx']
        self.logy          = graph_options['logy']

        self.xmin          = graph_options['xmin']
        self.xmax          = graph_options['xmax']
        self.xdiv          = graph_options['xdiv']
        self.ymin          = graph_options['ymin']
        self.ymax          = graph_options['ymax']
        self.ydiv          = graph_options['ydiv']
        self.xmin          = graph_options['xmin']
        self.xmax          = graph_options['xmax']
        self.xdiv          = graph_options['xdiv']
        self.ymin          = graph_options['ymin']
        self.ymax          = graph_options['ymax']
        self.ydiv          = graph_options['ydiv']
        self.title         = graph_options['title']
        self.xaxis         = graph_options['xaxis']
        self.yaxis         = graph_options['yaxis']
        self.plot()

    def plot(self):
        self.displayBridge.set_tight_layout()
        self.displayBridge.clear_axis()
        self.displayBridge.switch_axes(hide_axes2 = True)
        self.displayBridge.grid = self.grid

        for i in range(len(self.Multi_Model.models)):
            if self.Multi_Model.arquivos[i]['marker'] == True:
                self.Plot_sx_sy(self.Multi_Model.dfs[i], self.Multi_Model.arquivos[i])
        self.displayBridge.set_axes_props_without_axes_2(self.xmin, self.xmax, self.xdiv,
         self.ymin, self.ymax, self.ydiv, self.grid, self.logx, self.logy)
        left, right = self.displayBridge.axes1.get_xlim()
        self.displayBridge.axes1.set_xlim(left = left, right = right)
        self.displayBridge.axes1.set_xlim(left = left, right = right)
        lines = list()
        for i in range(len(self.Multi_Model.models)):
            if self.Multi_Model.arquivos[i]['func'] == True and self.Multi_Model.models[i] != 0.:
                self.Func_plot(self.Multi_Model.arquivos[i], self.Multi_Model.models[i], self.Multi_Model.parameters[i], self.Multi_Model.indVars[i], left, right, lines)
        self.displayBridge.axes1.set_title(self.title,
                                        fontsize=self.displayBridge.font_sizes["titulo"])
        self.displayBridge.axes1.set_xlabel(xlabel=self.xaxis,
                                fontsize=self.displayBridge.font_sizes["eixo_x"])
        self.displayBridge.axes1.set_ylabel(ylabel=self.yaxis,
                                fontsize=self.displayBridge.font_sizes["eixo_y"])
        handles, labels = self.displayBridge.axes1.get_legend_handles_labels()

        if len(handles) > 1:
            labels.reverse()
            handles.reverse()
            by_label = dict(zip(labels, handles))
            self.displayBridge.axes1.legend(by_label.values(), by_label.keys(),
            frameon=False,
            fontsize = self.displayBridge.font_sizes["legenda"],
            loc = self.displayBridge.legend_loc)
            
        elif len(handles) == 1:
            by_label = dict(zip(labels, handles))
            self.displayBridge.axes1.legend(by_label.values(), by_label.keys(),
            frameon=False,
            fontsize = self.displayBridge.font_sizes["legenda"],
            loc = self.displayBridge.legend_loc)
        self.displayBridge.canvas.draw_idle()
        

    def Plot_sx_sy(self, df, options):
        if options['label'] != '':
            self.displayBridge.axes1.errorbar(df['x'], df['y'], yerr=df['sy'], xerr = df['sx'],
            ecolor = options['markerColor'], capsize = 0, elinewidth = 1, ms = 3, marker = '.', color = options['markerColor'], ls = 'none', label = options['label'])
        else:
            self.displayBridge.axes1.errorbar(df['x'], df['y'], yerr=df['sy'], xerr = df['sx'],
            ecolor = options['markerColor'], capsize = 0, elinewidth = 1, ms = 3, marker = '.', color = options['markerColor'], ls = 'none')
    
    def Plot_sy(self, df, options): # Depreciada
        if options['label'] != '':
            self.displayBridge.axes1.errorbar(df['x'], df['y'], yerr=df['sy'],
            ecolor = options['markerColor'], capsize = 0, elinewidth = 1, ms = 3, marker = '.', color = options['markerColor'], ls = 'none', label = options['label'])
        else:
            self.displayBridge.axes1.errorbar(df['x'], df['y'], yerr=df['sy'],
            ecolor = options['markerColor'], capsize = 0, elinewidth = 1, ms = 3, marker = '.', color = options['markerColor'], ls = 'none')
    
    def Plot_op(self, df, options): # Depreciada
        if options['label'] != '':
            self.displayBridge.axes1.errorbar(df['x'], df['y'],
            ecolor = options['markerColor'], capsize = 0, elinewidth = 1, ms = 3, marker = '.', color = options['markerColor'], ls = 'none', label = options['label'])
        else:
            self.displayBridge.axes1.errorbar(df['x'], df['y'],
            ecolor = options['markerColor'], capsize = 0, elinewidth = 1, ms = 3, marker = '.', color = options['markerColor'], ls = 'none')
    
    def Func_plot(self, options, model, params, var, left, right, lines):
        x_plot = None
        if self.logx:
            x_plot = np.logspace(np.np.log10(left), np.np.log10(right), int(self.displayBridge.axes1.figure.get_size_inches()[0]*self.displayBridge.axes1.figure.dpi*1.75))
        else:
            x_plot = np.linspace(left, right, int(self.displayBridge.axes1.figure.get_size_inches()[0]*self.displayBridge.axes1.figure.dpi*1.75)) 
        y_plot = eval("model.eval(%s = x_plot, params = params)"%var, None,
         {'x_plot': x_plot, 'model': model, 'params': params})
        if options['label'] != '':
            line_func, = self.displayBridge.axes1.plot(x_plot, y_plot, lw = 1.5, color = options['markerColor'], ls = options['curve'], label = options['label'])
            lines.append(line_func)
        else:
            line_func, = self.displayBridge.axes1.plot(x_plot, y_plot, lw = 1.5, color = options['markerColor'], ls = options['curve'])
            lines.append(line_func)



# -*- coding: utf-8 -*-
"""
MIT License

Copyright (c) 2021 Leonardo Eiji Tamayose, Guilherme Ferrari Fortino

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import numpy as np
from PyQt5.QtCore import QObject
from scipy.stats import chi2, norm, t
from numpy import array

def calc_chi2_sim(ngl, nc):
    return array([chi2.ppf(0.5 - nc/2, ngl), chi2.ppf(0.5 + nc/2, ngl)])

def calc_chi2_lim_inf(ngl, nc):
    return chi2.ppf(1 - nc, ngl)

def calc_chi2_lim_sup(ngl, nc):
    return chi2.ppf(nc, ngl)

def calc_gauss_sim(nc):
    return array([norm.ppf(0.5 - nc/2), norm.ppf(0.5 + nc/2)])

def calc_gauss_lim_inf(nc):
    return norm.ppf(1 - nc)

def calc_gauss_lim_sup(nc):
    return norm.ppf(nc)

def calc_t_sim(ngl, nc):
    return array([t.ppf(0.5 - nc/2, df = ngl), t.ppf(0.5 + nc/2, df = ngl)])

def calc_t_lim_inf(ngl, nc):
    return t.ppf(1 - nc, df = ngl)

def calc_t_lim_sup(ngl, nc):
    return t.ppf(nc, df = ngl)

def calc_chi2r_sim(ngl, nc):
    return [chi2.ppf(0.5 - nc/2, ngl), chi2.ppf(0.5 + nc/2, ngl)]

def calc_chi2r_lim_inf(ngl, nc):
    return chi2.ppf(1 - nc, ngl)

def calc_chi2r_lim_sup(ngl, nc):
    return chi2.ppf(nc, ngl)

def interpreter_calculator(f, opt, nc, ngl, mean, std):
    ''' Return the plot string and the arrays for the graph plot.
    '''
    lim_inf = 0.0001
    lim_sup = 0.9999
    if nc > 0.9999:
        dif = (1 - nc)/2
        lim_inf = dif
        lim_sup = nc + dif
    
    if f == 0:
        # If it's Chi²
        x_plot = np.linspace(chi2.ppf(lim_inf, ngl), chi2.ppf(lim_sup, ngl), 350)
        y_plot = chi2.pdf(x_plot, ngl)
        s      = f"Calculadora de intervalo de confiança do Chi quadrado.\nNível de confiança {nc}.\nNível de significância {1 - nc:.7}\n"

        if opt == 0:
            result = calc_chi2_sim(ngl, nc)
            s     += "Intervalo simétrico.\nLimite inferior = %f \n Limite superior  = %f"%(result[0], result[1])
            x_area = np.linspace(result[0], result[1], 350)
            y_area = chi2.pdf(x_area, ngl)
            return s, x_plot, y_plot, x_area, y_area, f"Chi² com NGL = {int(ngl)}", "Chi²", "Densidade de probabilidade"

        elif opt == 1:
            result = calc_chi2_lim_inf(ngl, nc)
            s     += "Apenas limite inferior.\nLimite inferior = %f \n Limite superior = inf"%result
            x_area = np.linspace(result, chi2.ppf(lim_sup, ngl), 350)
            y_area = chi2.pdf(x_area, ngl)
            return s, x_plot, y_plot, x_area, y_area, f"Chi² com NGL = {int(ngl)}", "Chi²", "Densidade de probabilidade" 

        result = calc_chi2_lim_sup(ngl, nc)
        s     += "Apenas limite superior.\nLimite inferior = -inf \n Limite superior = %f"%result
        x_area = np.linspace(chi2.ppf(lim_inf, ngl), result, 350)
        y_area = chi2.pdf(x_area, ngl)
        return s, x_plot, y_plot, x_area, y_area, f"Chi² com NGL = {int(ngl)}", "Chi²", "Densidade de probabilidade"

    elif f == 1:
        # If it's Red Chi²
        x_plot = np.linspace(chi2.ppf(lim_inf, ngl), chi2.ppf(lim_sup, ngl), 350)
        y_plot = chi2.pdf(x_plot, ngl)
        x_plot = x_plot/ngl
        y_plot = y_plot*ngl
        s      = f"Calculadora de intervalo de confiança do Chi quadrado reduzido.\nNível de confiança {nc}.\nNível de significância {1 - nc:.7}\n"

        if opt == 0:
            result = calc_chi2r_sim(ngl, nc)
            s     += "Intervalo simétrico.\nLimite inferior = %f \n Limite superior  = %f"%(result[0]/ngl, result[1]/ngl)
            x_area = np.linspace(result[0], result[1], 350)
            y_area = chi2.pdf(x_area, ngl)*ngl
            x_area = x_area/ngl
            return s, x_plot, y_plot, x_area, y_area, f"Chi² reduzido com NGL = {int(ngl)}", "Chi² reduzido", "Densidade de probabilidade"

        elif opt == 1:
            result = calc_chi2r_lim_inf(ngl, nc)
            s     += "Apenas limite inferior.\nLimite inferior = %f \n Limite superior = inf"%(result/ngl)
            x_area = np.linspace(result, chi2.ppf(lim_sup, ngl), 350)
            y_area = chi2.pdf(x_area, ngl)*ngl
            x_area = x_area/ngl
            return s, x_plot, y_plot, x_area, y_area, f"Chi² reduzido com NGL = {int(ngl)}", "Chi² reduzido", "Densidade de probabilidade"

        result = calc_chi2r_lim_sup(ngl, nc)
        s     += "Apenas limite superior.\nLimite inferior = -inf \n Limite superior = %f"%(result/ngl)
        x_area = np.linspace(chi2.ppf(lim_inf, ngl), result, 350)
        y_area = chi2.pdf(x_area, ngl)*ngl
        x_area = x_area/ngl
        return s, x_plot, y_plot, x_area, y_area, f"Chi² reduzido com NGL = {int(ngl)}", "Chi² reduzido", "Densidade de probabilidade"

    elif f == 2:
        # If it's Gaussian
        x_plot = np.linspace(norm.ppf(lim_inf) , norm.ppf(lim_sup), 350)
        y_plot = norm.pdf(x_plot)
        x_plot = (x_plot * std) + mean
        s      = f"Calculadora de intervalo de confiança de Gaussiana.\nNível de confiança {nc}.\nNível de significância {1 - nc:.7}\n"

        if opt == 0:
            result = calc_gauss_sim(nc)
            s     += "Intervalo simétrico.\nLimite inferior = %f \n Limite superior  = %f"%(result[0]*std + mean, result[1]*std + mean)
            x_area = np.linspace(result[0], result[1], 350)
            y_area = norm.pdf(x_area)
            x_area = x_area*std + mean
            return s, x_plot, y_plot/std, x_area, y_area/std, "Gaussiana", "Valor", "Densidade de probabilidade"

        elif opt == 1:
            result = calc_gauss_lim_inf(nc)
            s     += "Apenas limite inferior.\nLimite inferior = %f \n Limite superior = inf"%(result*std + mean)
            x_area = np.linspace(result, norm.ppf(lim_sup), 350)
            y_area = norm.pdf(x_area)
            x_area = x_area*std + mean
            return s, x_plot, y_plot/std, x_area, y_area/std, "Gaussiana", "Valor", "Densidade de probabilidade"

        result = calc_gauss_lim_sup(nc)
        s     += "Apenas limite superior.\nLimite inferior = -inf \n Limite superior = %f"%(result*std + mean)
        x_area = np.linspace(result, norm.ppf(lim_inf), 350)
        y_area = norm.pdf(x_area)
        x_area = x_area*std + mean
        return s, x_plot, y_plot/std, x_area, y_area/std, "Gaussiana", "Valor", "Densidade de probabilidade"

    # If it's Student:    
    x_plot = np.linspace(t.ppf(lim_inf, df = ngl), t.ppf(lim_sup, df = ngl), 350)
    y_plot = t.pdf(x_plot, df = ngl)
    x_plot = x_plot * std + mean
    s      = f"Calculadora de intervalo de confiança de Student.\nNível de confiança {nc} | Nível de significância {1 - nc:.7}\n"

    if opt == 0:
        result = calc_t_sim(ngl, nc)
        s     += "Intervalo simétrico.\nLimite inferior = %f \n Limite superior  = %f"%(result[0]*std + mean, result[1]*std + mean)
        x_area = np.linspace(result[0], result[1], 350)
        y_area = t.pdf(x_area, df = ngl)
        x_area = x_area*std + mean
        return s, x_plot, y_plot/std, x_area, y_area/std, "Student", "Valor", "Densidade de probabilidade"
    
    elif opt == 1:
            result = calc_t_lim_inf(ngl, nc)
            s     += "Apenas limite inferior.\nLimite inferior = %f \n Limite superior = inf"%(result*std + mean)
            x_area = np.linspace(result, t.ppf(lim_sup, df = ngl), 350)
            y_area = t.pdf(x_area, df = ngl)
            x_area = x_area*std + mean
            return s, x_plot, y_plot/std, x_area, y_area/std, "Student", "Valor", "Densidade de probabilidade"

    result = calc_t_lim_sup(ngl, nc)
    s     += "Apenas limite superior.\nLimite inferior = -inf \n Limite superior = %f"%(result*std + mean)
    x_area = np.linspace(result, t.ppf(lim_inf, df = ngl), 350)
    y_area = t.pdf(x_area, df = ngl)
    x_area = x_area*std + mean
    return s, x_plot, y_plot/std, x_area, y_area/std, "Student", "Valor", "Densidade de probabilidade"

class CalculatorCanvas(QObject):
    """ A bridge class to interact with the plot in python
    """
    def __init__(self, parent=None):
        super().__init__(parent)
        # self.figure  = None
        

def plot(displayBridge, x, y, x_area, y_area, title, xlabel, ylabel):
    displayBridge.set_tight_layout()
    displayBridge.clear_axis()
    displayBridge.switch_axes(hide_axes2 = True)
    displayBridge.grid = True
    displayBridge.axes1.grid(True)
    displayBridge.axes1.fill_between(x_area, y_area, color = 'blue', alpha = 0.3)
    displayBridge.axes1.plot(x, y, lw = 1, c = 'red')
    displayBridge.axes1.set_title("P.D.F. " + title,
     fontsize= displayBridge.font_sizes["titulo"])
    displayBridge.axes1.set_xlabel(xlabel = xlabel,
     fontsize=displayBridge.font_sizes["eixo_x"])
    displayBridge.axes1.set_ylabel(ylabel = ylabel,
     fontsize=displayBridge.font_sizes["eixo_y"])
    displayBridge.set_axes_props_without_axes_2(None, None, None, None, None, None, None, False, False)
    displayBridge.canvas.draw_idle()


# -*- coding: utf-8 -*-
"""
MIT License

Copyright (c) 2021 Leonardo Eiji Tamayose, Guilherme Ferrari Fortino

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from PyQt5.QtCore import QObject, QJsonValue, QUrl, pyqtSignal, pyqtSlot
from matplotlib.ticker import MaxNLocator
import pandas as pd
import numpy as np
import platform
import json

class Histogram(QObject):
    """Backend for histogram page"""

    # Signals
    fillPage = pyqtSignal(QJsonValue)

    def __init__(self, canvas, messageHandler) -> None:
        super().__init__()
        self.messageHandler = messageHandler
        self.canvas         = canvas
        self.path           = ""
        self.histAlign      = {"Centro" : "mid", "Direita" : "right", "Esquerda" : "left"}
        self.histOrient     = {"Vertical" : "vertical", "Horizontal" : "horizontal"}

    @pyqtSlot(QJsonValue)
    def plot(self, data):
        """TODO: Docstring for histogram plot."""
        data = QJsonValue.toVariant(data)
        self.canvas.set_tight_layout()
        self.canvas.clear_axis()
        self.canvas.switch_axes(hide_axes2 = True)

        xdiv = data["props"]["xdiv"]
        xmin = data["props"]["xmin"]
        xmax = data["props"]["xmax"]
        ydiv = data["props"]["ydiv"]
        ymin = data["props"]["ymin"]
        ymax = data["props"]["ymax"]

        self.canvas.grid = data["props"]["grid"]

        self.canvas.axes1.yaxis.set_major_locator(MaxNLocator(integer=True))

        has_legend = False
        for arquivo in data["data"]:
            if arquivo["visible"]:
                if data["props"]["histMode"] == "Frequência absoluta":
                    has_legend = self.plot_freq_abs(arquivo, data, has_legend)
                elif data["props"]["histMode"] == "Frequência relativa":
                    has_legend = self.plot_freq_rel(arquivo, data, has_legend)
                else:
                    has_legend = self.plot_dens(arquivo, data, has_legend)
        self.canvas.axes1.set_title(data["props"]["title"], 
                                fontsize=self.canvas.font_sizes["titulo"])
        self.canvas.axes1.set_xlabel(xlabel = data["props"]["xaxis"],
                                fontsize=self.canvas.font_sizes["eixo_x"])
        self.canvas.axes1.set_ylabel(ylabel = data["props"]["yaxis"],
                                fontsize=self.canvas.font_sizes["eixo_y"])
        self.canvas.set_axes_props_without_axes_2(xmin, xmax, xdiv, ymin, ymax, ydiv,
         data["props"]["grid"], data["props"]["logx"], data["props"]["logy"])
        if has_legend:
            self.canvas.axes1.legend()
        self.canvas.canvas.draw_idle()

    def plot_freq_abs(self, arquivo, data, has_legend):
        df    = pd.DataFrame.from_dict(json.loads(arquivo["data"]))
        alpha = self.make_float(arquivo["kargs"].pop("alpha"), 1.0)
        label = arquivo["kargs"].pop("label")
        left  = self.make_float(arquivo["kargs"].pop("rangexmin", ""), df["x"].min())
        right = self.make_float(arquivo["kargs"].pop("rangexmax", ""), df["x"].max())
        if left >= right:
            self.messageHandler.raise_error("Intervalo de bins inválido. Rever intervalo de bins.")
            return -1
        bins  = np.linspace(left, right,
            self.make_int(arquivo["kargs"].pop("nbins", 10), 10) + 1)
        counts = None
        if arquivo["legend"] == "":
            counts, bins, _ = self.canvas.axes1.hist(x = df["x"], bins = bins,
            density     = False, cumulative = False, bottom = 0,
            histtype    = data["props"]["histType"],
            align       = self.histAlign[data["props"]["histAlign"]],
            orientation = self.histOrient[data["props"]["histOrientation"]],
            log = False, rwidth = 1, capstyle = "round", ls = "-", aa = True,
            alpha = alpha,
            **arquivo["kargs"])
        else:
            counts, bins, _ = self.canvas.axes1.hist(x = df["x"], bins = bins,
            density     = False, cumulative = False, bottom = 0,
            histtype    = data["props"]["histType"],
            align       = self.histAlign[data["props"]["histAlign"]],
            orientation = self.histOrient[data["props"]["histOrientation"]],
            log = False, rwidth = 1, capstyle = "round", ls = "-", aa = True,
            alpha = alpha, label = arquivo["legend"],
            **arquivo["kargs"])
            has_legend = True
        if label:
            if data["props"]["histOrientation"] == "Vertical":
                bottom, top = self.canvas.axes1.get_ylim()
                height = top - bottom
                c = (bins[1] - bins[0])/2
                for n, b in zip(counts, bins):
                    if n != 0:
                        self.canvas.axes1.text(b + c, n + height*0.02, str(n), ha = "center")
            else:
                left, right = self.canvas.axes1.get_xlim()
                dif = right - left
                c = (bins[1] - bins[0])/2
                for n, b in zip(counts, bins):
                    if n != 0:
                        self.canvas.axes1.text(n + dif*0.02, b + c, str(n), ha = "center")
        return has_legend

    def plot_freq_rel(self, arquivo, data, has_legend):
        df    = pd.DataFrame.from_dict(json.loads(arquivo["data"]))
        alpha = self.make_float(arquivo["kargs"].pop("alpha"), 1.0)
        label = arquivo["kargs"].pop("label")
        left  = self.make_float(arquivo["kargs"].pop("rangexmin", ""), df["x"].min())
        right = self.make_float(arquivo["kargs"].pop("rangexmax", ""), df["x"].max())
        if left >= right:
            self.messageHandler.raise_error("Intervalo de bins inválido. Rever intervalo de bins.")
            return -1
        bins  = np.linspace(left, right,
            self.make_int(arquivo["kargs"].pop("nbins", 10), 10) + 1)
        counts = None
        if arquivo["legend"] == "":
            counts, bins, _ = self.canvas.axes1.hist(x = df["x"], bins = bins, weights = np.ones_like(df["x"])/len(df["x"]),
            density     = False, cumulative = False, bottom = 0,
            histtype    = data["props"]["histType"],
            align       = self.histAlign[data["props"]["histAlign"]],
            orientation = self.histOrient[data["props"]["histOrientation"]],
            log = False, rwidth = 1, capstyle = "round", ls = "-", aa = True,
            alpha = alpha,
            **arquivo["kargs"])
        else:
            counts, bins, _ = self.canvas.axes1.hist(x = df["x"], bins = bins, weights = np.ones_like(df["x"])/len(df["x"]),
            density     = False, cumulative = False, bottom = 0,
            histtype    = data["props"]["histType"],
            align       = self.histAlign[data["props"]["histAlign"]],
            orientation = self.histOrient[data["props"]["histOrientation"]],
            log = False, rwidth = 1, capstyle = "round", ls = "-", aa = True,
            alpha = alpha, label = arquivo["legend"],
            **arquivo["kargs"])
            has_legend = True
        if label:
            if data["props"]["histOrientation"] == "Vertical":
                bottom, top = self.canvas.axes1.get_ylim()
                height = top - bottom
                c = (bins[1] - bins[0])/2
                for n, b in zip(counts, bins):
                    if n != 0:
                        self.canvas.axes1.text(b + c, n + height*0.02, f"{n:.3g}", ha = "center")
            else:
                left, right = self.canvas.axes1.get_xlim()
                dif = right - left
                c = (bins[1] - bins[0])/2
                for n, b in zip(counts, bins):
                    if n != 0:
                        self.canvas.axes1.text(n + dif*0.02, b + c, f"{n:.3g}", ha = "center")
        return has_legend

    def plot_dens(self, arquivo, data, has_legend):
        df    = pd.DataFrame.from_dict(json.loads(arquivo["data"]))
        alpha = self.make_float(arquivo["kargs"].pop("alpha"), 1.0)
        label = arquivo["kargs"].pop("label")
        left  = self.make_float(arquivo["kargs"].pop("rangexmin", ""), df["x"].min())
        right = self.make_float(arquivo["kargs"].pop("rangexmax", ""), df["x"].max())
        if left >= right:
            self.messageHandler.raise_error("Intervalo de bins inválido. Rever intervalo de bins.")
            return -1
        bins  = np.linspace(left, right,
            self.make_int(arquivo["kargs"].pop("nbins", 10), 10) + 1)
        counts = None
        if arquivo["legend"] == "":
            counts, bins, _ = self.canvas.axes1.hist(x = df["x"], bins = bins,
            density     = True, cumulative = False, bottom = 0,
            histtype    = data["props"]["histType"],
            align       = self.histAlign[data["props"]["histAlign"]],
            orientation = self.histOrient[data["props"]["histOrientation"]],
            log = False, rwidth = 1, capstyle = "round", ls = "-", aa = True,
            alpha = alpha,
            **arquivo["kargs"])
        else:
            counts, bins, _ = self.canvas.axes1.hist(x = df["x"], bins = bins,
            density     = True, cumulative = False, bottom = 0,
            histtype    = data["props"]["histType"],
            align       = self.histAlign[data["props"]["histAlign"]],
            orientation = self.histOrient[data["props"]["histOrientation"]],
            log = False, rwidth = 1, capstyle = "round", ls = "-", aa = True,
            alpha = alpha, label = arquivo["legend"],
            **arquivo["kargs"])
            has_legend = True
        if label:
            if data["props"]["histOrientation"] == "Vertical":
                bottom, top = self.canvas.axes1.get_ylim()
                height = top - bottom
                c = (bins[1] - bins[0])/2
                for n, b in zip(counts, bins):
                    if n != 0:
                        self.canvas.axes1.text(b + c, n + height*0.02, f"{n:.3g}", ha = "center")
            else:
                left, right = self.canvas.axes1.get_xlim()
                dif = right - left
                c = (bins[1] - bins[0])/2
                for n, b in zip(counts, bins):
                    if n != 0:
                        self.canvas.axes1.text(n + dif*0.02, b + c, f"{n:.3g}", ha = "center")
        return has_legend

    @pyqtSlot()
    def new(self):
        self.path = ""
        self.fillPage.emit(None)
        self.canvas.clear_axis()
        self.canvas.switch_axes(True)

    @pyqtSlot(QJsonValue, result=bool)
    def save(self, data):
        # If there's no path for saving, save_as()
        if self.path == '':
            return False

        # Getting properties
        data = data.toVariant()

        if platform.system() == "Linux":
            if self.path[-5:] == ".json":
                with open(self.path, 'w', encoding='utf-8') as file:
                    json.dump(data, file, ensure_ascii=False, indent=4)
            else: 
                with open(self.path + ".json", 'w', encoding='utf-8') as file:
                    json.dump(data, file, ensure_ascii=False, indent=4)
        else:
            with open(self.path, 'w', encoding='utf-8') as file:
                json.dump(data, file, ensure_ascii=False, indent=4)

        return True

    @pyqtSlot(str, QJsonValue)
    def save_as(self, path, data):
        # Getting path
        self.path = QUrl(path).toLocalFile()
        self.save(data)

    @pyqtSlot(str)
    def load(self, path):
        # Reseting frontend
        self.new()

        # Getting path
        self.path = QUrl(path).toLocalFile()

        # Getting props
        with open(self.path, encoding='utf-8') as file:
            props = json.load(file)

        if "key" in props:
            if props["key"].split('-')[-1] != 'hist':
                self.messageHandler.raise_error("Este projeto pertence à outra aba do ATUS.")
                return 0
            if props["key"][0] == "2":
                self.messageHandler.raise_warn("Arquivo da versão anterior. Procure salvar o arquivo nesta versão para evitar problemas.")
        else:
            self.messageHandler.raise_error("O arquivo carregado é incompatível com o ATUS.")
            return 0

        self.fillPage.emit(QJsonValue.fromVariant(props))

    @pyqtSlot(str, result=QJsonValue)
    def check_data(self, filePath):
        '''
        Check if data is valid
        Returns: True + Data if valid data, False otherwise
        '''
        package = {
            "isValid": False,
            "data"   : None
        }
        # Loading from .csv or (.txt and .tsv)
        filePath = QUrl(filePath).toLocalFile()
        if filePath[-3:] == "csv":
            try:
                df = pd.read_csv(filePath, sep=',', header=None, dtype = str).replace(np.nan, "0")
            except pd.errors.ParserError:
                self.messageHandler.raise_error("Separação de colunas de arquivos csv são com vírgula (","). Rever dados de entrada.")
                # Separação de colunas de arquivos csv são com vírgula (","). Rever dados de entrada.
                return QJsonValue.fromVariant(package)
        elif filePath[-3:] == "tsv" or filePath[-3:] == "txt":
            try:
                df = pd.read_csv(filePath, sep='\t', header=None, dtype = str).replace(np.nan, "0")
            except pd.errors.ParserError:
                self.messageHandler.raise_error("Separação de colunas de arquivos txt e tsv são com tab. Rever dados de entrada.")
                # Separação de colunas de arquivos txt e tsv são com tab. Rever dados de entrada.
                return QJsonValue.fromVariant(package)
        else:
            self.messageHandler.raise_error("Apenas arquivos .txt, .csv, .tsv são suportados.")
            return QJsonValue.fromVariant(package)

        if len(df.columns) == 1:
            df.columns = ["x"]
        else:
            self.messageHandler.raise_error("A tabela de histogramas deve conter no máximo 1 coluna.")
            return QJsonValue.fromVariant(package)
        
        for i in df.columns:
            # Replacing comma for dots
            df[i] = [x.replace(',', '.') for x in df[i]]
            # Converting everything to float
            try:
                df[i] = df[i].astype(float)
            except:
                self.messageHandler.raise_error("A entrada de dados só permite entrada de números. Rever arquivo de entrada.")
                # Há células não numéricas. A entrada de dados só permite entrada de números. Rever arquivo de entrada.
                return QJsonValue.fromVariant(package)

        package["data"] = df.to_json()
        package["isValid"] = True
        return package

    def make_float(self, var, value):
        try:
            return float(var)
        except:
            return value

    def make_int(self, var, value):
        try:
            return int(var)
        except:
            return value


# -*- coding: utf-8 -*-
"""
MIT License

Copyright (c) 2021 Leonardo Eiji Tamayose, Guilherme Ferrari Fortino

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import print_function
import os.path
# from matplotlib_backend_qtquick.qt_compat import QtCore
from PyQt5.QtCore import QObject, pyqtProperty, QUrl, pyqtSignal, pyqtSlot, QJsonValue
# from googleapiclient.discovery import build
# from googleapiclient.http import MediaFileUpload
# from google_auth_oauthlib.flow import InstalledAppFlow
# from google.auth.transport.requests import Request
# from google.oauth2.credentials import Credentials

class GDrive(QObject):
    '''Class that interact with google drive API'''
    # Signals
    informationSignal = pyqtSignal(QJsonValue, arguments='ownersInfo')

    def __init__(self, messageHandler):
        super().__init__()

        # Paths
        self.credsPath = os.path.join(os.path.abspath(__file__), '../../credentials.json')
        self.tokenPath = os.path.join(os.path.expanduser('~/Documents'), 'token.json')

        # print(self.credsPath)
        # print(self.tokenPath)
        # print(os.path.exists(self.credsPath))
        # print(os.path.exists(self.tokenPath))
        
        # Properties
        self.messageHandler = messageHandler
        self.creds = None
        self.service = None

        # If modifying these scopes, delete the file token.json.
        self.SCOPES = ['https://www.googleapis.com/auth/drive.metadata.readonly',
                       'https://www.googleapis.com/auth/drive.appdata',
                       'https://www.googleapis.com/auth/userinfo.profile']

    # @pyqtSlot()
    # def tryLogin(self):
    #     '''
    #         Called when the program is initialized, try to connect with Google if there's already a token
    #     '''
    #     # If there's already a token, get creds
    #     if os.path.exists(self.tokenPath):
    #         self.creds = Credentials.from_authorized_user_file(self.tokenPath, self.SCOPES)
    #         about = self.getDrive()
    #         self.informationSignal.emit(QJsonValue.fromVariant(about))

    # @pyqtSlot()
    # def login(self):
    #     '''
    #         Connect to Google Drive API and get user's information
    #     '''
    #     # If there are no (valid) credentials available, let the user log in.
    #     if not self.creds or not self.creds.valid:
    #         if self.creds and self.creds.expired and self.creds.refresh_token:
    #             try:
    #                 self.creds.refresh(Request())
    #             except:
    #                 return 0
    #         else:
    #             flow = InstalledAppFlow.from_client_secrets_file('credentials.json', self.SCOPES)
    #             try:
    #                 self.creds = flow.run_local_server(port=0)
    #             except:
    #                 return 0

    #         # Save the credentials for the next run
    #         with open(self.tokenPath, 'w') as token:
    #             token.write(self.creds.to_json())
            
    #         # Mounting drive
    #         about = self.getDrive()
    #         self.informationSignal.emit(QJsonValue.fromVariant(about))
    
    # @pyqtSlot()
    # def logout(self):
    #     '''
    #         Closes http2lib connection
    #     '''
    #     self.service.close()

    #     # Removing token
    #     os.remove(self.tokenPath)
    #     self.service = None
    #     self.creds = None

    # @pyqtSlot()
    # def listFiles(self):
    #     results = self.service.files().list(spaces='appDataFolder', fields="nextPageToken, files(id, name)", q="mimeType='application/json'").execute()
    #     print(results.get('files', []))
    #     # return results.get('files', [])

    # @pyqtSlot()
    # def uploadFile(self):
    #     file_metadata = {
    #         'name': 'config.json',
    #         'parents': ['appDataFolder']
    #     }
    #     media = MediaFileUpload(self.tokenPath,
    #                             mimetype='application/json',
    #                             resumable=True)
    #     file = self.service.files().create(body=file_metadata,
    #                                         media_body=media,
    #                                         fields='id').execute()
    #     print('File ID: %s' % file.get('id'))

    # def getDrive(self):
    #     '''
    #         Build Google Drive API
    #         Returns: information about the drive's owner
    #     '''
    #     # Building drive
    #     self.service = build('drive', 'v3', credentials=self.creds)

    #     # Getting information about drive's owner
    #     about = self.service.about().get(fields='user').execute()

    #     return about


from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open('atus/version.txt') as version:
    __VERSION__  = version.read()

setup(
    name="atus",
    version=__VERSION__,
    author="High Elo Devs",
    author_email="atusdevs@gmail.com",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/HighEloDevs/Analysis-Tool-for-Undergrad-Students",
    project_urls={
        "Bug Tracker":
        "https://github.com/HighEloDevs/Analysis-Tool-for-Undergrad-Students/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_data={
        'atus': ['version.txt', 'icon.ico'],
    },
    package_dir={"src": "src"},
    packages=["atus", "atus/src"],
    include_package_data=True,
    python_requires=">=3.7, <3.10",
    entry_points={
        'console_scripts': ['atus = atus:main.main'],
    },
    install_requires=[
        "setuptools>=42",
        "wheel",
        "matplotlib==3.4.2",
        "pandas",
        "numpy",
        "scipy",
        "lmfit",
        "matplotlib_backend_qtquick",
        "pyqt5",
        "requests",
        "pytexit",
    ])



# -*- coding: utf-8 -*-
"""
MIT License

Copyright (c) 2021 Leonardo Eiji Tamayose, Guilherme Ferrari Fortino

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import os

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import gridspec
from matplotlib_backend_qtquick.backend_qtquick import NavigationToolbar2QtQuick
from PyQt5.QtCore import QObject, QUrl, pyqtProperty, pyqtSignal, pyqtSlot
from PyQt5.QtGui import QGuiApplication, QPixmap


class Canvas(QObject):
    """TODO"""

    # Some signals for the frontend
    coordinates_changed = pyqtSignal(str)

    def __init__(self, message_handler):
        super().__init__()

        self.message_handler = message_handler

        # The figure, canvas, toolbar and axes
        self.figure = None
        self.canvas = None
        self.toolbar = None
        self.axes = None
        self.axes1 = None
        self.axes2 = None
        self.oid = 0
        self.cid = 0
        self.grid = False
        self.top = 0.92
        self.bottom = 0.12
        self.left = 0.10
        self.right = 0.95
        self.figmode = 0
        self.legend_loc_dict = {
            "Automático" : 0,
            "Direita-Superior" : 1,
            "Esquerda-Superior" : 2,
            "Esquerda-Inferior" : 3,
            "Direita-Inferior" : 4,
            "Esquerda-Centro" : 6,
            "Direita-Centro" : 7,
            "Centro-Inferior" : 8,
            "Centro-Superior" : 9,
            "Centro-Centro" : 10
        }
        self.legend_loc = "best"
        self.dpi = 500
        self.font_sizes = {
            "titulo" : 12,
            "residuos" : 12,
            "legenda" : 12,
            "eixo_x" : 12,
            "eixo_y" : 12,
        }
        self.user_alpha_outliers = 0.25
        self.user_color_outliers = None
        # This is used to display the coordinates of the mouse in the window
        self._coordinates = ""

    def update_with_canvas(self, canvas):
        """Initialize with the canvas for the figure."""
        self.canvas = canvas
        self.figure = self.canvas.figure
        self.toolbar = NavigationToolbar2QtQuick(canvas=canvas)
        self.gm = gridspec.GridSpec(2,
                                    1,
                                    figure=self.figure,
                                    height_ratios=[3.0, 1.0])
        self.gs = gridspec.GridSpec(1,
                                    1,
                                    figure=self.figure,
                                    height_ratios=[1.0])
        self.axes1 = self.figure.add_subplot(self.gm[0],
                                             picker=True,
                                             autoscale_on=True)
        self.axes2 = self.figure.add_subplot(self.gm[1],
                                             picker=True,
                                             sharex=self.axes1,
                                             autoscale_on=False)
        self.set_tight_layout()
        self.axes2.set_visible(False)
        self.axes1.set_position(self.gs[:, :].get_position(self.figure),
                                which="original")
        self.axes1.grid(False)
        self.canvas.draw_idle()

        # Connect for displaying the coordinates
        self.figure.canvas.mpl_connect("motion_notify_event", self.on_motion)

    def clear_axis(self):
        """Clear the current plot in the axis."""
        self.axes1.cla()
        self.axes2.cla()
        self.axes1.relim()
        self.axes2.relim()
        self.axes1.remove_callback(self.oid)
        self.axes1.figure.canvas.mpl_disconnect(self.cid)
        self.canvas.draw_idle()

    def switch_axes(self, hide_axes2: bool = True):
        """Função que oculta ou não o eixo secundário."""
        if hide_axes2:
            self.axes2.set_visible(False)
            self.axes1.set_position(self.gs[0].get_position(self.figure),
                                    which="both")
            self.figmode = 0
        else:
            self.axes2.set_visible(True)
            self.axes1.set_position(self.gm[0].get_position(self.figure),
                                    which="both")
            self.axes2.set_position(self.gm[1].get_position(self.figure),
                                    which="both")
            self.figmode = 1

    def set_axes_props_without_axes_2(self, xmin, xmax, xdiv, ymin, ymax, ydiv,
                                      grid, log_x, log_y):
        left, right = self.axes1.get_xlim()
        bottom, top = self.axes1.get_ylim()
        divs_x = len(self.axes1.get_xticks()) - 1
        divs_y = len(self.axes1.get_yticks()) - 1
        xmin = self.make_float(xmin, left)
        xmax = self.make_float(xmax, right)
        xdiv = self.make_int(xdiv, divs_x)
        ymin = self.make_float(ymin, bottom)
        ymax = self.make_float(ymax, top)
        ydiv = self.make_int(ydiv, divs_y)

        if grid:
            self.axes1.grid(True, which="major")
        if log_y:
            self.axes1.set_yscale("log")
        if log_x:
            self.axes1.set_xscale("log")

        if xdiv != divs_x:
            self.axes1.set_xticks(np.linspace(xmin, xmax, xdiv + 1))
            self.axes1.set_xlim(left=xmin, right=xmax)
        else:
            if left != xmin or right != xmax:
                self.axes1.set_xticks(np.linspace(xmin, xmax, xdiv + 1))
                self.axes1.set_xlim(left=xmin, right=xmax)

        if ydiv != divs_y:
            self.axes1.set_yticks(np.linspace(ymin, ymax, ydiv + 1))
            self.axes1.set_ylim(bottom=ymin, top=ymax)
        else:
            if bottom != ymin or top != ymax:
                self.axes1.set_yticks(np.linspace(ymin, ymax, ydiv + 1))
                self.axes1.set_ylim(bottom=ymin, top=ymax)

    def set_axes_props_with_axes_2(self, xmin, xmax, xdiv, ymin, ymax, ydiv,
                                   resmin, resmax, grid, log_x, log_y):
        left, right = self.axes1.get_xlim()
        bottom, top = self.axes1.get_ylim()
        botres, topres = self.axes2.get_ylim()
        divs_x = len(self.axes1.get_xticks()) - 1
        divs_y = len(self.axes1.get_yticks()) - 1

        xmin = self.make_float(xmin, left)
        xmax = self.make_float(xmax, right)
        xdiv = self.make_int(xdiv, divs_x)
        ymin = self.make_float(ymin, bottom)
        ymax = self.make_float(ymax, top)
        ydiv = self.make_int(ydiv, divs_y)
        self.axes2.set_ylim(bottom=self.make_float(resmin, botres),
                            top=self.make_float(resmax, topres))
        if xdiv != divs_x:
            self.axes1.set_xticks(np.linspace(xmin, xmax, xdiv + 1))
            self.axes2.set_xticks(np.linspace(xmin, xmax, xdiv + 1))
            self.axes1.set_xlim(left=xmin, right=xmax)
            self.axes2.set_xlim(left=xmin, right=xmax)
        else:
            if left != xmin or right != xmax:
                self.axes1.set_xticks(np.linspace(xmin, xmax, xdiv + 1))
                self.axes2.set_xticks(np.linspace(xmin, xmax, xdiv + 1))
                self.axes1.set_xlim(left=xmin, right=xmax)
                self.axes2.set_xlim(left=xmin, right=xmax)
            else:
                self.axes1.set_xlim(left=xmin, right=xmax)
                self.axes2.set_xlim(left=xmin, right=xmax)

        if ydiv != divs_y:
            self.axes1.set_yticks(np.linspace(ymin, ymax, ydiv + 1))
            self.axes1.set_ylim(bottom=ymin, top=ymax)
        else:
            if bottom != ymin or top != ymax:
                self.axes1.set_yticks(np.linspace(ymin, ymax, ydiv + 1))
                self.axes1.set_ylim(bottom=ymin, top=ymax)

        if grid:
            self.axes1.grid(True, which="major")
            self.axes2.grid(True, which="major")
        if log_y:
            self.axes1.set_yscale("log")
        if log_x:
            self.axes1.set_xscale("log")
        plt.setp(self.axes1.get_xticklabels(), visible=False)

    def plot_error_bar(self, x, y, kargs_errorbar, alpha, sy= None, sx = None, y_r = None, ssy = None):
        self.axes1.errorbar(x, y, yerr=sy, xerr=sx, alpha = alpha, **kargs_errorbar)
        if y_r is not None:
            self.axes2.errorbar(x, y_r, yerr=ssy, alpha = alpha, **kargs_errorbar)

    def set_tight_layout(self):
        self.figure.subplots_adjust(left=self.left,
                                    bottom=self.bottom,
                                    right=self.right,
                                    top=self.top)

    def make_float(self, var, valor):
        try:
            return float(var)
        except:
            return valor

    def make_int(self, var, valor):
        try:
            return int(var)
        except:
            return valor

    def get_coordinates(self):
        """Retorna as coordenadas no gráfico."""
        return self._coordinates

    def set_coordinates(self, coordinates):
        """Seta as coordenadas do gráfico."""
        self._coordinates = coordinates
        self.coordinates_changed.emit(self._coordinates)

    @pyqtSlot(str, str, str, str)
    def set_paddings(self, top, bottom, left, right):
        self.top = self.make_float(top, valor=0.92)
        self.bottom = self.make_float(bottom, valor=0.12)
        self.left = self.make_float(left, valor=0.10)
        self.right = self.make_float(right, valor=0.95)
        if self.bottom >= self.top or self.left >= self.right:
            self.message_handler.raise_error("Paddings invalidos.")
            return None
        if self.figmode:
            self.figure.subplots_adjust(left=self.left,
                                        bottom=self.bottom,
                                        right=self.right,
                                        top=self.top)
        else:
            self.figure.subplots_adjust(left=self.left,
                                        bottom=self.bottom,
                                        right=self.right,
                                        top=self.top)
            self.switch_axes()
        self.canvas.draw_idle()

    @pyqtSlot(result=str)
    def get_paddings(self):
        return "0.92;0.12;0.10;0.95"

    @pyqtSlot(str, bool)
    def save_plot(self, save_path, transparent):
        """Gets the path from input and save the actual plot."""
        path = QUrl(save_path).toLocalFile()

        # Getting extension
        _, extension = os.path.splitext(path)  # Recebe filename e extension

        if transparent and extension != "png":
            self.message_handler.raise_warn(
                "O fundo transparente funciona apenas na extensão .png")
            self.canvas.figure.savefig(path, dpi = self.dpi, transparent = transparent)
        else:
            self.canvas.figure.savefig(path, dpi = self.dpi, transparent = transparent)
            self.message_handler.raise_success("Imagem salva com sucesso!")

    @pyqtSlot()
    def copy_to_clipboard(self):
        """Copy imagine to the clipboard."""
        # Getting clipboard
        clipboard = QGuiApplication.clipboard()

        # Saving image to a path
        try:
            path = os.path.join(os.path.expanduser(r"~\Documents"),
                                "image.png")
            self.canvas.figure.savefig(path, dpi=150, transparent=False)
            pixmap = QPixmap()
            # Loading image as pixmap and saving to clipboard
            if pixmap.load(path):
                clipboard.setImage(pixmap.toImage())
                self.message_handler.raise_success(
                    "Copiado com sucesso para a área de transferência!")
            os.remove(path)
        except:
            self.message_handler.raise_error(
                "Erro copiar para a área de transferência, contatar os desenvolvedores."
            )

    @pyqtSlot()
    def pan(self, *args):
        self.toolbar.pan(*args)

    @pyqtSlot()
    def zoom(self, *args):
        self.toolbar.zoom(*args)

    @pyqtSlot()
    def home(self, *args):
        self.toolbar.home(*args)

    @pyqtSlot()
    def back(self, *args):
        self.toolbar.back(*args)

    @pyqtSlot()
    def forward(self, *args):
        self.toolbar.forward(*args)

    @pyqtSlot()
    def shortcut_grid(self):
        self.axes1.grid(not self.grid)
        self.axes2.grid(not self.grid)
        self.grid = not self.grid
        self.canvas.draw_idle()

    @pyqtSlot()
    def shortcut_axis_1(self):
        if self.axes1.axison:
            self.axes1.axis("off")
        else:
            self.axes1.axis("on")
        self.canvas.draw_idle()

    @pyqtSlot()
    def shortcut_axis_2(self):
        if self.axes2.axison:
            self.axes2.axis("off")
        else:
            self.axes2.axis("on")
        self.canvas.draw_idle()

    @pyqtSlot(int, int)
    def set_canvas_size(self, width, height):
        if width == 0 or height == 0:
            self.resize_canvas()
            return
        dpi = self.canvas.figure.get_dpi()
        self.canvas.figure.set_size_inches(width / dpi, height / dpi)
        self.canvas.draw_idle()

    @pyqtSlot(result=list)
    def get_canvas_size(self):
        # Getting the dpi of the figure
        dpi = self.canvas.figure.get_dpi()
        # Getting the size of the figure
        width, height = self.canvas.figure.get_size_inches()

        return [int(width * dpi), int(height * dpi), dpi]

    @pyqtSlot()
    def resize_canvas(self):
        """Resizes the figure to fit the canvas"""
        self.canvas.geometryChanged(self.canvas.boundingRect(),
                                    self.canvas.boundingRect())

    @pyqtSlot(int, int, int, int, int)
    def set_font_sizes(self, title_size, x_size, y_size, residual_size,
                       caption_size):
        self.font_sizes["titulo"] = title_size
        self.font_sizes["eixo_x"] = x_size
        self.font_sizes["eixo_y"] = y_size
        self.font_sizes["residuos"] = residual_size
        self.font_sizes["legenda"] = caption_size
        if self.figmode:
            self.axes1.set_title(self.axes1.get_title(), fontsize=self.font_sizes["titulo"])
            self.axes1.set_ylabel(self.axes1.get_ylabel(), fontsize=self.font_sizes["eixo_y"])
            self.axes2.set_xlabel(self.axes2.get_xlabel(), fontsize=self.font_sizes["eixo_x"])
            self.axes2.set_ylabel(self.axes2.get_ylabel(), fontsize=self.font_sizes["residuos"])
        else:
            self.axes1.set_title(self.axes1.get_title(), fontsize=self.font_sizes["titulo"])
            self.axes1.set_xlabel(self.axes1.get_xlabel(), fontsize=self.font_sizes["eixo_x"])
            self.axes1.set_ylabel(self.axes1.get_ylabel(), fontsize=self.font_sizes["eixo_y"])
        self.canvas.draw_idle()

    @pyqtSlot(str)
    def set_legend_position(self, position):
        self.legend_loc = self.legend_loc_dict[position]
        if self.axes1.get_legend():
            h, l = self.axes1.get_legend_handles_labels()
            self.axes1.legend(h, l, loc=self.legend_loc,
                 fontsize=self.font_sizes["legenda"])

    @pyqtSlot(int)
    def set_dpi(self, dpi):
        self.dpi = dpi

    @pyqtSlot(float)
    def set_opacity_outliers(self, value:int):
        self.user_alpha_outliers = value

    def on_motion(self, event):
        """Update the coordinates on the display."""
        if event.inaxes in (self.axes1, self.axes2):
            self.coordinates = f"({event.xdata:.2f}, {event.ydata:.2f})"

    coordinates = pyqtProperty(str,
                               get_coordinates,
                               set_coordinates,
                               notify=coordinates_changed)
